"""Implementation of EcModel class.

Extend the XbaModel to become an enzyme constraint model.
The implementation of GECKO is based on the GECKO 1.0 matlab package (Sánchez et al., 2017).
The implementation of ccFBA and MOMENTmr is based on the R-package sybilccFBA (Desouki, 2015)
The implementation of MOMENT is based on literature (Adadi et al., 2012)

Peter Schubert, HHU Duesseldorf, October 2023
"""

import re
import numpy as np
import pandas as pd
import f2xba.prefixes as pf
from ..utils.mapping_utils import valid_sbml_sid, load_parameter_file

MAX_CONC_PROT = 100  # mg/gDW maximum protein concentration in kpmf (kilo protein mass fraction)


class EcModel:
    """Extend a XbaModel to an enzyme constraint model.

    The implementation of GECKO is based on the GECKO 1.0 matlab package (Sánchez et al., 2017),
    ccFBA and MOMENTmr on the R-package sybilccFBA (Desouki, 2015) and MOMENT on literature
    (Adadi et al., 2012).

    .. code-block:: python

        xba_model = XbaModel('iJO1366.xml')
        xba_model.configure('iJO1366_GECKO_xba_parameters.xlsx')

        ec_model = EcModel(xba_model)
        ec_model.configure('iJO1366_GECKO_ecm_parameters.xlsx')
        if ec_model.validate():
            ec_model.export('iJO1366_GECKO.xml')
    """

    def __init__(self, xba_model):
        """Instantiate the EcModel instance.

        :param xba_model: a reference to a XbaModel instance
        :type xba_model: :class:`f2xba.XbaModel`
        """
        self.model = xba_model
        self.arm_flag = False

    def configure(self, fname):
        """Configuration with ECM configuration data.

        Accepted table: 'general'

        :param str fname: filename of ECM configuration file (.xlsx)
        :return: success
        :rtype: bool
        """
        sheet_names = ['general', 'rescale_reactions']
        ecm_params = load_parameter_file(fname, sheet_names)
        if 'general' not in ecm_params.keys():
            print(f'mandatory table "general" not found in the document')
            raise ValueError

        general_params = ecm_params['general']['value'].to_dict()
        ecm_type = general_params.get('ecm_type', 'GECKO')
        p_total = general_params.get('p_total', 0.5)
        avg_enz_sat = general_params.get('avg_enz_sat', 0.5)
        self.arm_flag = general_params.get('arm_flag', False)

        if 'pm2totpm' in general_params:
            pm2totpm = general_params['pm2totpm']
        else:
            assert 'pm2totpm_val_or_paxdb' in general_params, 'parameter "pm2totpm_val_or_paxdb" is required'
            if type(general_params['pm2totpm_val_or_paxdb']) is float:
                pm2totpm = general_params['pm2totpm_val_or_paxdb']
            else:
                pm2totpm = self.get_modelled_protein_mass_fraction(general_params['pm2totpm_val_or_paxdb'])
        print(f'modeled protein fraction of total protein mass {pm2totpm:.4f} g/g')

        if 'rescale_reactions' in ecm_params:
            self._rescale_reactions(ecm_params['rescale_reactions'])

        # create enzyme constraint model
        if ecm_type in ['GECKO', 'MOMENTmr']:
            self.model.add_isoenzyme_reactions()
            self.model.make_irreversible()
            self._add_gecko_protein_species()
        elif ecm_type == 'MOMENT':
            self.model.add_isoenzyme_reactions()
            self._add_moment_protein_species()
            self.model.make_irreversible()
        elif ecm_type == 'ccFBA':
            self._select_ccfba_isoenzyme()
            self.model.make_irreversible()
            self.model.remove_unused_gps()
            self._add_gecko_protein_species()
        else:
            print(f'model not constructed, invalid ecm_type {ecm_type}')
            return False

        if self.arm_flag is True:
            self.model.add_arm_reactions()

        protein_pool = p_total * pm2totpm
        print(f'protein constraint: {protein_pool*1000.0:.2f} mg/gDW - modeled protein')
        self._add_total_protein_constraint(protein_pool)
        self._reaction_enzyme_coupling(avg_enz_sat)

        # remove model components that are not used in the model
        self.model.clean()

        # add some parameter values for reference (after model.clean() so they are not removed)
        add_params = {'gmodeledP_per_gP': {'value': pm2totpm, 'name': 'protein mass fraction modeled'},
                      'gP_per_gDW': {'value': p_total, 'name': 'protein mass fraction of total dry mass'},
                      'avg_enz_sat': {'value': avg_enz_sat, 'name': 'average enzyme saturation level'}}
        for pid, p_data in add_params.items():
            self.model.add_parameter(pid, p_data)

        # modify some model attributs and create L3V2 SBML model
        self.model.model_attrs['id'] += f'_{ecm_type}'
        if 'name' in self.model.model_attrs:
            self.model.model_attrs['name'] = f'{ecm_type} model of ' + self.model.model_attrs['name']
        self.model.sbml_container['level'] = 3
        self.model.sbml_container['version'] = 2

        self.model.print_size()
        return True

    def _get_enzyme_gene_products(self):
        """Determine used genes products for enzymes.

        Note: Model can contain additional gene products for
        gene product associations where enzyme kcat have not
        been provided. Useful for gene deletion studies.

        :return: gene product ids used by the enzymes in the model
        :rtype: set of str
        """
        used_enzymes = set()
        for rid, r in self.model.reactions.items():
            for eid in r.enzymes:
                used_enzymes.add(eid)
        used_enz_gps = set()
        for eid in used_enzymes:
            e = self.model.enzymes[eid]
            for gene_id in e.composition:
                used_enz_gps.add(self.model.locus2gp[gene_id])
        return used_enz_gps

    def _add_gecko_protein_species(self):
        """Add protein species to the GECKO/ccFBA/MOMENTmr model.

        Note: Protein compartments in XBA model are derived from
        reaction compartments, which in turn is a concatenation of
        species compartments being substrates to the reaction.

        Here we only need one of the species compartments.

        We only create proteins when gene product is used by any of the
        enzymes

        add protein species to the model (also ensure that protein species name is valid)
        add protein drain reactions to the model
        """
        used_enz_gps = self._get_enzyme_gene_products()

        protein_sids = {}
        for gpid in used_enz_gps:
            uid = self.model.gps[gpid].uid
            p = self.model.proteins[uid]
            prot_sid = pf.C_prot_ + valid_sbml_sid(uid)
            p.link_sid(prot_sid)
            prot_metaid = f'meta_prot_{valid_sbml_sid(uid)}'
            prot_annot = f'bqbiol:is, uniprot/{uid}'
            protein_sids[prot_sid] = [p.name, p.cid, False, False, False, prot_metaid, prot_annot]

        cols = ['name', 'compartment', 'hasOnlySubstanceUnits', 'boundaryCondition',
                'constant', 'metaid', 'miriamAnnotation']
        df_add_species = pd.DataFrame(protein_sids.values(), index=list(protein_sids), columns=cols)
        print(f'{len(df_add_species):4d} protein constraints to add')
        self.model.add_species(df_add_species)

    def _add_moment_protein_species(self):
        """Add protein species to MOMENT model.

        Execute before making model irreversilbe.

        In MOMENT promiscuous enzymes can catalyze several reactions in parallel
        To implement this constraint, we add for each catalzyed (iso)reaction
        a specific protein species

        add protein species to the model
        """
        # add dedicated protein species for each (iso)reaction
        protein_sids = {}
        for rid, r in self.model.reactions.items():
            assert(len(r.enzymes) <= 1)
            if len(r.enzymes) == 1:
                enz = self.model.enzymes[r.enzymes[0]]
                for locus in enz.composition:
                    uid = self.model.locus2uid[locus]
                    p = self.model.proteins[uid]
                    prot_sid = pf.C_prot_ + valid_sbml_sid(uid) + f'_{rid}'
                    p.link_sid(prot_sid)
                    prot_metaid = f'meta_prot_{valid_sbml_sid(uid)}_{rid}'
                    prot_annot = f'bqbiol:is, uniprot/{uid}'
                    protein_sids[prot_sid] = [p.name, p.cid, False, False, False, prot_metaid, prot_annot]

        cols = ['name', 'compartment', 'hasOnlySubstanceUnits', 'boundaryCondition',
                'constant', 'metaid', 'miriamAnnotation']
        df_add_species = pd.DataFrame(protein_sids.values(), index=list(protein_sids), columns=cols)
        print(f'{len(df_add_species):4d} moment protein constraints to add')
        self.model.add_species(df_add_species)

    def _select_ccfba_isoenzyme(self):
        """Select least cost enzyme for reactions catalyzed by several isoenzymes.

        Consider number of active sites when determining enzyme kcat value
        Other isoenzymes get deleted and gpa updated.

        Integration with Thermodynamic FBA (TFA):
            TFA splits TD related reations in fwd/rev (_REV)
            TFA also split reactions that in base model are not reversible
            i.e. we may have irreversible '<rid>_REV' with no kcat

        Selection is based on molecular weight and turnover number.
        Select enzyme with minimal MW/kcat.

        Different isoenzymes might be selected for forward/reverse direction.
        Therefore, reversible reactions with isoenzymes get split in fwd/bwd reaction.
        """
        # Note model.reactions get updated within the loop, therefore fix rids initially

        rids = list(self.model.reactions)
        for rid in rids:
            r = self.model.reactions[rid]
            if len(r.enzymes) >= 2 and r.kcatsf is not None and np.isfinite(r.kcatsf[0]):
                directional_rs = [r]
                if r.reversible:
                    # split reaction, i.e. create a new irreversible reverse reaction
                    rev_r = self.model.split_reversible_reaction(r)
                    directional_rs.append(rev_r)

                for dir_r in directional_rs:
                    valid_kcatsf = (sum([1 for kcat in dir_r.kcatsf if np.isfinite(kcat)])
                                    if dir_r.kcatsf is not None else 0)
                    # consider reactions with undefined kcatf (due to TD integration)
                    if len(r.enzymes) == valid_kcatsf:
                        idx = np.argmin([self.model.enzymes[eid].mw / (kcat * self.model.enzymes[eid].active_sites)
                                         for eid, kcat in zip(dir_r.enzymes, dir_r.kcatsf)])
                        eid = dir_r.enzymes[idx]
                        gpa = ' and '.join(sorted([self.model.uid2gp[self.model.locus2uid[locus]]
                                                   for locus in self.model.enzymes[eid].composition]))
                        if ' and ' in gpa:
                            gpa = '(' + gpa + ')'
                        dir_r.enzymes = [dir_r.enzymes[idx]]
                        dir_r.kcatsf = [dir_r.kcatsf[idx]]
                        dir_r.gene_product_assoc = gpa
                    else:
                        dir_r.enzymes = []
                        dir_r.kcatsf = None
                        dir_r.kcatsr = None

    def _add_total_protein_constraint(self, total_protein):
        """add total modeled protein constraint to the enzyme constraint model.

        add protein pool species to the model
        add protein drain reactions to the model
        add protein pool exchange reation to the model

        protein concentration in model has units of mg/gDW (i.e. 1000.0 * pmf)

        :param float total_protein: total modeled protein mass in g/gDW
        """
        # add total protein pool species in default compartment
        pool_sid = f'{pf.C_prot_pool}'
        pool_name = 'total protein pool'
        cols = ['name', 'compartment', 'hasOnlySubstanceUnits', 'boundaryCondition', 'constant']
        df_add_species = pd.DataFrame([[pool_name, self.model.main_cid, False, False, False]],
                                      index=[pool_sid], columns=cols)
        self.model.add_species(df_add_species)

        # add total protein pool exchange reaction
        protein_vars = {}
        draw_rid = pf.V_PC_total
        draw_name = f'total concentration modeled protein'
        zero_mg_pid = self.model.get_fbc_bnd_pid(0.0, 'mg_per_gDW', 'zero_mass_conc_mg_per_gDW')
        total_prot_mg_pid = self.model.get_fbc_bnd_pid(total_protein * 1000.0, 'mg_per_gDW', 'conc_protein_mg_per_gDW')
        protein_vars[draw_rid] = [draw_name, f' => {pool_sid}', zero_mg_pid, total_prot_mg_pid, None, 'protein']

        # add protein drain reactions - supporting MOMENT with specific protein split in protein species per reaction
        max_conc_mg_pid = self.model.get_fbc_bnd_pid(MAX_CONC_PROT, 'mg_per_gDW', 'max_conc_prot_mg_per_gDW')

        # create only for used proteins
        used_enz_gps = self._get_enzyme_gene_products()
        for gpid in used_enz_gps:
            uid = self.model.gps[gpid].uid
            p = self.model.proteins[uid]
            draw_rid = pf.V_PC_ + valid_sbml_sid(uid)
            draw_name = f'conc_prot_{uid}'
            # supporting MOMENT model with protein split into protein species per reaction
            products = ' + '.join([sid for sid in p.linked_sids])
            # reaction_string = f'{p.mw / 1000.0} {pool_sid} => {products}'
            reaction_string = f'{pool_sid} => {products}'
            protein_vars[draw_rid] = [draw_name, reaction_string, zero_mg_pid, max_conc_mg_pid,
                                      self.model.uid2gp[uid], 'protein']

        cols = ['name', 'reactionString', 'fbcLowerFluxBound', 'fbcUpperFluxBound', 'fbcGeneProdAssoc', 'kind']
        df_add_rids = pd.DataFrame(protein_vars.values(), index=list(protein_vars), columns=cols)
        print(f'{len(df_add_rids):4d} protein variables to add')
        self.model.add_reactions(df_add_rids)

    def get_modelled_protein_mass_fraction(self, fname):
        """Determine protein mass fraction based on Pax-db.org downlaod.

        Determine which part of organism protein mass fraction is modelled.
        Protein abundance file needs first to be downloaded from Pax-db.org

        :param str fname: file path/name of protein abundance data collected from Pax-db.org
        :return: relative pmf of model based
        :rtype: float
        """
        # parse PacDb file
        name = ''
        publication = ''
        last_comment_line = ''
        gene_idx = None
        ppm_abundance = {}
        with open(fname) as file:
            for line in file:
                line = line.rstrip()
                if re.match('#', line):
                    if re.match('#name: ', line):
                        name = line.split(': ')[1]
                    elif re.match('#publication_year: ', line):
                        publication = line.split(': ')[1]
                    last_comment_line = line
                else:
                    if gene_idx is None:
                        cols = re.sub('^#', '', last_comment_line).split('\t')
                        gene_idx = cols.index('string_external_id')
                        ppm_idx = cols.index('abundance')
                    record = line.split('\t')
                    gene = record[gene_idx].split('.')[1]
                    ppm_abundance[gene] = float(record[ppm_idx])
        print(f'PaxDb file {name}, year {publication}, covering {len(ppm_abundance)} proteins '
              f'(total {sum(ppm_abundance.values()):.1f} ppm) loaded from {fname}')

        used_enz_gps = self._get_enzyme_gene_products()
        used_uids = {self.model.gps[gpid].uid for gpid in used_enz_gps}

        p_rel_total = 0.0
        p_rel_model = 0.0
        for locus, ppm in ppm_abundance.items():
            if locus in self.model.uniprot_data.locus2uid:
                uid = self.model.uniprot_data.locus2uid[locus]
                rel_mass = ppm / 1.0e6 * self.model.uniprot_data.proteins[uid].mass
                p_rel_total += rel_mass
                if uid in used_uids:
                    p_rel_model += rel_mass
        return p_rel_model / p_rel_total

    def _reaction_enzyme_coupling(self, avg_enz_sat):
        """Couple reactions with enzyme/protein requirement via kcats.

        applicable to enzyme catalyzed reactions.
        from reaction get enzyme and corresponding kcat
        convert kcat from s-1 to h-1
        scaled total enzyme kcat by number of active sites and average enzyme saturation

        protein constraint for a specific protein:
          reaction flux [mmol/gDWh] ≤ kcat [1/s] * 3600 [s/h] * n_AS * 1/stoic * avg_enz_sat * P_conc [mg/gDW]
                                      * 1/MW [mmol/mg]
          - stoic: number of protein copies in the enzyme
          - n_AS: number of active sites
          - MW: protein molecular weight in Da = g/mol = mg/mmol
        Except for MOMENT we assume equality to optimize protein allocation
          constraint per protein: C_prot_uid couples reaction flux R_xxx with protein concentration V_PC_uid
          C_prot_uid: (-1/(kcat * 3600 * n_AS * avg_enz_sat) * stoic * MW) * R_xxx + 1 * V_PC_uid = (≤) 0

        :param float avg_enz_sat: average enzyme saturation applicable to all reactions
        """
        for r in self.model.reactions.values():
            assert len(r.enzymes) <= 1, ('Something went wrong, at this stage of isoreactions, we expect a maximum'
                                         'of one enzyme catalyzing this reaction')
            if len(r.enzymes) == 1 and r.kcat is not None:
                e = self.model.enzymes[r.enzymes[0]]
                enz_kcat_per_h = r.kcat * 3600.0 * e.active_sites * avg_enz_sat

                for locus, stoic in e.composition.items():
                    uid = self.model.locus2uid[locus]
                    p = self.model.proteins[uid]
                    prot_sid = None
                    # in MOMENT model, a promiscuous enzyme/protein has specific proteins per reactions
                    # e.g. b2215, ompC, P06996: this protein is linked to all (reversible) reactions it catalyzes
                    #  reactions R_23CCMPtex_iso4 and R_23CCMPtex_iso4_REV both coupled to constraint
                    #  C_prot_P06996_R_23CAMPtex_iso4
                    if len(p.linked_sids) == 1:
                        prot_sid = list(p.linked_sids)[0]
                    else:
                        rev_rid = re.sub('_REV$', '', r.id)
                        for linked_sid in p.linked_sids:
                            if rev_rid in linked_sid:
                                prot_sid = linked_sid
                                break
                    r.reactants[prot_sid] = (stoic / enz_kcat_per_h) * p.mw

    def _rescale_reactions(self, df_rescale):
        """Rescale reactions (e.g. Biomass) as per ecYeast7 (Sanchez, 2017)

        biomass dict contains
        - biomass reaction id 'rid'
        - rescale tasks 'rescale' a list of dict, each with
            - either rescale factor 'factor' or an absolute value 'value', flaat
            - 'rectants' and/or 'products', consisting of list of species ids

        amino acids get scaled by parameter f_p
        carbohydrates get scaled by parameter f_c
        GAM related species get set to gam level

        :param df_rescale: biomass rescaling parameters
        :type df_rescale: pandas.DataFrame
        :return:
        """
        modify_reactants = {}
        modify_products = {}
        for rid, row in df_rescale.iterrows():
            r = self.model.reactions[rid]
            if type(row['reactants']) is str:
                scale_reacs = {sid.strip() for sid in row['reactants'].split(',')}
                if rid not in modify_reactants:
                    modify_reactants[rid] = r.reactants.copy()
                new_srefs = {}
                if np.isfinite(row['factor']):
                    factor = row['factor']
                    for sid, stoic in modify_reactants[rid].items():
                        new_srefs[sid] = factor * stoic if sid in scale_reacs else stoic
                elif np.isfinite(row['value']):
                    value = row['value']
                    for sid, stoic in modify_reactants[rid].items():
                        new_srefs[sid] = value if sid in scale_reacs else stoic
                modify_reactants[rid] = new_srefs

            if type(row['products']) is str:
                scale_prods = {sid.strip() for sid in row['products'].split(',')}
                if rid not in modify_products:
                    modify_products[rid] = r.products.copy()
                new_srefs = {}
                if np.isfinite(row['factor']):
                    factor = row['factor']
                    for sid, stoic in modify_products[rid].items():
                        new_srefs[sid] = factor * stoic if sid in scale_prods else stoic
                elif np.isfinite(row['value']):
                    value = row['value']
                    for sid, stoic in modify_products[rid].items():
                        new_srefs[sid] = value if sid in scale_prods else stoic
                modify_products[rid] = new_srefs

        modify_attrs = []
        for rid, srefs in modify_reactants.items():
            reactants = '; '.join([f'species={sid}, stoic={stoic}' for sid, stoic in srefs.items()])
            modify_attrs.append([rid, 'reaction', 'reactants', reactants])
        for rid, srefs in modify_products.items():
            products = '; '.join([f'species={sid}, stoic={stoic}' for sid, stoic in srefs.items()])
            modify_attrs.append([rid, 'reaction', 'products', products])
        cols = ['id', 'component', 'attribute', 'value']
        df_modify_attrs = pd.DataFrame(modify_attrs, columns=cols)
        df_modify_attrs.set_index('id', inplace=True)
        print(f'reaction rescaling')
        self.model.modify_attributes(df_modify_attrs, 'reaction')

    def query_kcats(self):
        """Query enzyme-reaction kcat values.

        Generate a table that can be used to configure kcats

        :return: table with kcat values
        :rtype: pandas.DataFrame
        """
        notes = f'kcat export from {self.model.model_attrs.id}'
        rid_kcats = []
        for sub_rid, r in self.model.reactions.items():
            tmp_rid = re.sub('_REV$', '', sub_rid)
            rid = re.sub(r'_iso\d$', '', tmp_rid)
            if len(r.enzymes) == 1:
                dirxn = -1 if re.search('_REV$', sub_rid) else 1
                eid = r.enzymes[0]
                genes = ', '.join([locus.strip() for locus in eid.split('_')[1:]])
                active_sites = self.model.enzymes[eid].active_sites,
                rid_kcats.append([rid, sub_rid, dirxn, genes, r.kcatsf, active_sites, notes])
        cols = ['rid', 'sub_rid', 'dirxn', 'genes', 'kcat_per_s', 'active_sites', 'notes']
        df_rid_kcats = pd.DataFrame(rid_kcats, columns=cols)
        df_rid_kcats.sort_values(by=['sub_rid', 'dirxn'], inplace=True)
        df_rid_kcats.set_index('rid', inplace=True)
        return df_rid_kcats

    def validate(self):
        """Validate compliance to SBML standards.

        :return: success
        :rtype: bool
        """
        return self.model.validate()

    def export(self, fname):
        """Export EcModel to SBML coded file or in tabular format.

        :param str fname: filename (with extension '.xml' or '.xlsx')
        :return: success
        :rtype: bool
        """
        return self.model.export(fname)



"""Implementation of RbaProcesses, RbaProcess and RbaProcessingMap classes.

Peter Schubert, CCB, HHU Duesseldorf, December 2022
"""

import os
import re
import numpy as np
import pandas as pd
from xml.etree.ElementTree import parse, ElementTree, Element, SubElement, indent

from .rba_target_value import RbaTargetValue
from ..utils.rba_utils import get_species_refs_from_xml, get_species_refs_from_str, \
                              extract_params, translate_reaction_string


class RbaProcesses:

    def __init__(self):
        self.processes = {}
        self.processing_maps = {}

    def import_xml(self, model_dir):

        file_name = os.path.join(model_dir, 'processes.xml')
        if os.path.exists(file_name) is True:
            root = parse(file_name).getroot()
            assert root.tag == 'RBAProcesses'
            self.processes = RbaProcess.import_xml(root.find('listOfProcesses'))
            self.processing_maps = RbaProcessingMap.import_xml(root.find('listOfProcessingMaps'))
        else:
            print(f'{file_name} not found!')

    def from_df(self, m_dict):
        if 'processes' in m_dict:
            self.processes = RbaProcess.from_df(m_dict['processes'])
        else:
            print(f'processes not imported!')
        if 'processingMaps' in m_dict:
            self.processing_maps = RbaProcessingMap.from_df(m_dict['processingMaps'])
        else:
            print(f'processingMaps not imported!')

    def from_xba(self, rba_params, rba_model):
        """Configure Processes and Processing maps.

        Parameter functions and aggregates are created
        Cofactors are automatically added to translation map.

        'input_filter' for set 'rna'
        - comma-separated list of regular expression patterns that will be checked against the macromolecule id
            - e.g. 'mrna' to select only mrna
            - e.g. 'trna, b' to select only tRNAs and E. coli gene loci (used for rRNA)

        'input_filter' for set 'protein':
        - 'signal_peptide': select proteins having a signal peptide as per UniProt data
          plus dummy_proteins not located in the cytoplasm.
        - comma-separated list of valid RBA compartment ids (select proteins accordingly)
        - comma-separated list of regular expression patterns that will be checked against protein id

        :param rba_params: RBA model specific parametrization
        :type rba_params: dict of pandas DataFrames
        :param rba_model: rba model instance
        :type rba_model: class RbaModel
        """
        df_pmaps_data = rba_params['processing_maps']
        df_proc_data = rba_params['processes']
        df_mach_data = rba_params['machineries']
        cytoplasm_cid = rba_model.cid_mappings['cytoplasm_cid']

        # Process data
        pmap2set = {}
        for proc_id, row in df_proc_data.iterrows():
            proc_name = row['name']
            proc_set = row['set']
            proc_type = row['type']
            proc_map = row['processing_map']
            pmap2set[proc_map] = proc_set

            # add machinery, if defined
            constant = row.get('capacity_constant', np.nan)
            func = row.get('capacity_function', np.nan)
            agg = row.get('capacity_aggregate', np.nan)
            p_name = rba_model.parameters.create_parameter(f'{proc_id}_capacity', constant, func, agg)
            machinery = {}
            if p_name:
                machinery['capacity'] = RbaTargetValue.get_target_value('value', p_name)
                machinery['reactants'] = {}
                machinery['products'] = {}
                # note: df_mach_data.loc[[proc_id]] with double angular brackets to enforce dataframe being returned
                for _, mach_comp in df_mach_data.loc[[proc_id]].iterrows():
                    if mach_comp['stoic'] < 0.0:
                        machinery['reactants'][mach_comp['id']] = -mach_comp['stoic']
                    else:
                        machinery['products'][mach_comp['id']] = mach_comp['stoic']
            # determine process inputs
            inputs = []
            if proc_set == 'dna':
                inputs = list(rba_model.dna.macromolecules)

            elif proc_set == 'rna':
                if type(row['input_filter']) is not str:
                    inputs = list(rba_model.rnas.macromolecules)
                else:
                    patterns = [pattern.strip() for pattern in row['input_filter'].split(',')]
                    input_set = set()
                    for rnaid in rba_model.rnas.macromolecules:
                        for pattern in patterns:
                            if re.match(pattern, rnaid):
                                input_set.add(rnaid)
                                break
                    inputs = list(input_set)

            elif proc_set == 'protein':
                if type(row['input_filter']) is not str:
                    inputs = list(rba_model.proteins.macromolecules)
                elif row['input_filter'] == 'signal_peptide':
                    inputs = []
                    for locus in rba_model.proteins.macromolecules:
                        # add those proteins to the input, that have a signal_peptide in their UniProt data
                        if locus in rba_model.model.locus2uid:
                            p = rba_model.model.proteins[rba_model.model.locus2uid[locus]]
                            if p.has_signal_peptide is True:
                                inputs.append(locus)
                        # dummy proteins have no UniProt entries. We add them to input if they are located in cytoplasm
                        else:
                            if rba_model.proteins.macromolecules[locus].compartment != cytoplasm_cid:
                                inputs.append(locus)
                else:
                    items = {item.strip() for item in row['input_filter'].split(',')}
                    input_set = set()
                    # check if items are valid compartment ids:
                    if len(items.intersection(set(rba_model.metabolism.compartments))) == len(items):
                        for mm_id, mm in rba_model.proteins.macromolecules.items():
                            if mm.compartment in items:
                                input_set.add(mm_id)
                    else:
                        # consider items as patterns or names of macromolecules
                        for mm_id in rba_model.proteins.macromolecules:
                            for pattern in items:
                                if re.match(pattern, mm_id):
                                    input_set.add(mm_id)
                                    break
                    inputs = list(input_set)

            proc_data = {'processingMap': proc_map, 'set': proc_set, 'inputs': inputs}
            if proc_type == 'production':
                self.processes[proc_id] = RbaProcess(proc_id, name=proc_name, machinery=machinery,
                                                     productions=proc_data)
            elif proc_type == 'degradation':
                self.processes[proc_id] = RbaProcess(proc_id, name=proc_name, machinery=machinery,
                                                     degradations=proc_data)

        # processing maps
        set2components = {'dna': rba_model.dna.components, 'rna': rba_model.rnas.components,
                          'protein': rba_model.proteins.components}
        for pmapid, proc_set in pmap2set.items():
            const_proc = {}
            comp_proc = {}
            for _, row in df_pmaps_data.loc[[pmapid]].iterrows():
                mach_cost = row['machinery_cost']
                component = row['component']
                reactants, products = translate_reaction_string(row['reaction_string'])

                # constant processing data for macromolecule
                if component == 'constantProcessing':
                    const_proc['reactants'] = reactants
                    const_proc['products'] = products

                # comonent processing costs
                elif component in set2components[proc_set]:
                    # processing of a main component, e.g. a specific amino acid during translation
                    comp_proc[component] = {'cost': mach_cost, 'reactants': reactants, 'products': products}
                else:
                    # automatic creation of a set of component processings of specific type, e.g. folding of amino acids
                    for comp_id, comp in set2components[proc_set].items():
                        if comp.type == component:
                            if comp.type == 'cofactor':
                                # cofactor processing is based on cofactor consumption
                                comp_proc[comp_id] = {'cost': mach_cost, 'reactants': {comp_id: 1.0}, 'products': {}}
                            else:
                                # e.g. amino acid processing with same reaction string (e.g. folding, secretion)
                                comp_proc[comp_id] = {'cost': mach_cost, 'reactants': reactants, 'products': products}
                # configuration if individual reaction

            self.processing_maps[pmapid] = RbaProcessingMap(pmapid, const_processing=const_proc,
                                                            comp_processing=comp_proc)

        print(f'{len(self.processes):4d} processes and {len(self.processing_maps)} processing maps')

    def export_xml(self, model_dir):

        file_name = os.path.join(model_dir, 'processes.xml')
        root = Element('RBAProcesses')

        processes = SubElement(root, 'listOfProcesses')
        for item in self.processes.values():
            processes.append(item.export_xml())

        pmaps = SubElement(root, 'listOfProcessingMaps')
        for item in self.processing_maps.values():
            pmaps.append(item.export_xml())

        tree = ElementTree(root)
        indent(tree)
        tree.write(file_name)

    def to_df(self):
        df_p = pd.DataFrame([item.to_dict() for item in self.processes.values()])
        df_p.set_index('process', inplace=True)
        data = []
        for pmid, pmap in self.processing_maps.items():
            pm_dict = pmap.to_dict()
            if ((len(pm_dict['constantProcessing'].get('reactants', {})) > 0) or
                    (len(pm_dict['constantProcessing'].get('products', {})) > 0)):
                data.append([pmid, 'constantProcessing', np.nan, pm_dict['constantProcessing']['reactants'],
                             pm_dict['constantProcessing']['products']])
            for component, comp_proc in pm_dict['componentProcessings'].items():
                data.append([pmid, component, comp_proc['machineryCost'], comp_proc['reactants'],
                             comp_proc['products']])

        df_pm = pd.DataFrame(data, columns=['processingMap', 'component', 'machineryCost',
                                            'reactants', 'products'])
        df_pm.set_index('processingMap', inplace=True)

        return {'processes': df_p, 'processingMaps': df_pm}

    def validate(self, component_ids):
        valid = True

        missing = self.ref_molecules_pmaps().difference(component_ids['species'])
        if len(missing) > 0:
            print('species used in processingMaps not defined:', missing)
            valid = False

        missing = self.ref_molecules_machinery().difference(component_ids['species']) \
            .difference(component_ids['rna']) \
            .difference(component_ids['protein'])
        if len(missing) > 0:
            print('species/macromolecules used in processes not defined:', missing)
            valid = False

        missing = self.ref_molecules_inputs().difference(component_ids['dna']) \
            .difference(component_ids['rna']) \
            .difference(component_ids['protein'])
        if len(missing) > 0:
            print('macromolecules used in processes not defined:', missing)
            valid = False

        missing = self.ref_parameters().difference(component_ids['functions']) \
            .difference(component_ids['aggregates'])
        if len(missing) > 0:
            print('function/aggregates used in aggregates not defined:', missing)
            valid = False

        return valid

    def ref_molecules_pmaps(self):
        refs = set()
        for pmap in self.processing_maps.values():
            refs |= {sid for sid in pmap.constant_processing.get('reactants', {})}
            refs |= {sid for sid in pmap.constant_processing.get('products', {})}
            for comp_proc in pmap.component_processings.values():
                refs |= {sid for sid in comp_proc['reactants']}
                refs |= {sid for sid in comp_proc['products']}
        return refs

    def ref_molecules_machinery(self):
        refs = set()
        for p in self.processes.values():
            refs |= {sid for sid in p.machinery.get('reactants', {})}
            refs |= {sid for sid in p.machinery.get('products', {})}
        return refs

    def ref_molecules_inputs(self):
        refs = set()
        for p in self.processes.values():
            refs |= {sid for sid in p.productions.get('inputs', {})}
            refs |= {sid for sid in p.degradations.get('inputs', {})}
        return refs

    def ref_molecules(self):
        refs = set()
        refs |= self.ref_molecules_pmaps()
        refs |= self.ref_molecules_machinery()
        refs |= self.ref_molecules_inputs()
        return refs

    def ref_parameters(self):
        refs = set()
        for p in self.processes.values():
            if 'capacity' in p.machinery:
                refs.add(p.machinery['capacity'].value)
        return refs


class RbaProcess:

    def __init__(self, proc_id, name=None, machinery=None, productions=None, degradations=None):
        self.id = proc_id
        self.name = name
        self.machinery = machinery if type(machinery) is dict else {}
        self.productions = productions if type(productions) is dict else {}
        self.degradations = degradations if type(degradations) is dict else {}
        self.constr_id = None

    @staticmethod
    def import_xml(processes):

        data = {}
        for process in processes.findall('process'):
            pid = process.attrib['id']
            rba_process = RbaProcess(pid)
            rba_process.name = process.get('name')
            machinery = process.find('machinery')
            if machinery is not None:
                composition = machinery.find('machineryComposition')
                reactants = get_species_refs_from_xml(composition.find('listOfReactants'))
                products = get_species_refs_from_xml(composition.find('listOfProducts'))
                capacity = RbaTargetValue.from_dict(machinery.find('capacity').attrib)
                rba_process.machinery = {'capacity': capacity, 'reactants': reactants,
                                         'products': products}
            processings = process.find('processings')
            if processings is not None:
                productions = processings.find('listOfProductions')
                if productions is not None:
                    processing = productions.find('processing')
                    processing_map = processing.attrib['processingMap']
                    set_type = processing.attrib['set']
                    lo_inputs = processing.find('listOfInputs')
                    inputs = [sref.attrib['species'] for sref in lo_inputs.findall('speciesReference')]
                    rba_process.productions = {'processingMap': processing_map, 'set': set_type, 'inputs': inputs}
                degradations = processings.find('listOfDegradations')
                if degradations is not None:
                    processing = degradations.find('processing')
                    processing_map = processing.attrib['processingMap']
                    set_type = processing.attrib['set']
                    lo_inputs = processing.find('listOfInputs')
                    inputs = [sref.attrib['species'] for sref in lo_inputs.findall('speciesReference')]
                    rba_process.degradations = {'processingMap': processing_map, 'set': set_type, 'inputs': inputs}

            data[pid] = rba_process
        return data

    @staticmethod
    def from_df(df):
        data = {}
        for pid, row in df.iterrows():
            rba_process = RbaProcess(pid)
            rba_process.name = row['name']
            if type(row['machineryCapacity']) is str and len(row['machineryCapacity']) > 0:
                capacity_dict = extract_params(row['machineryCapacity'])
                capacity = RbaTargetValue.from_dict(capacity_dict)
                reactants = get_species_refs_from_str(row['machineryReactants'])
                products = get_species_refs_from_str(row['machineryProducts'])
                rba_process.machinery = {'capacity': capacity, 'reactants': reactants,
                                         'products': products}
            if type(row['productionProcessingMap']) is str and len(row['productionProcessingMap']) > 0:
                processing_map = row['productionProcessingMap']
                set_type = row['productionSet']
                inputs = [item.strip() for item in row['productionInputs'].split(',')]
                rba_process.productions = {'processingMap': processing_map, 'set': set_type, 'inputs': inputs}

            if type(row['degradationProcessingMap']) is str and len(row['degradationProcessingMap']) > 0:
                processing_map = row['degradationProcessingMap']
                set_type = row['degradationSet']
                inputs = [item.strip() for item in row['degradationInputs'].split(',')]
                rba_process.degradations = {'processingMap': processing_map, 'set': set_type, 'inputs': inputs}

            data[pid] = rba_process
        return data

    def export_xml(self):
        process = Element('process', {'id': self.id, 'name': self.name})

        if len(self.machinery) > 0:
            machinery = SubElement(process, 'machinery')
            composition = SubElement(machinery, 'machineryComposition')
            if len(self.machinery['reactants']) > 0:
                reactants = SubElement(composition, 'listOfReactants')
                for sid, stoic in self.machinery['reactants'].items():
                    SubElement(reactants, 'speciesReference', {'species': sid, 'stoichiometry': str(stoic)})
            if len(self.machinery['products']) > 0:
                products = SubElement(composition, 'listOfProducts')
                for sid, stoic in self.machinery['products'].items():
                    SubElement(products, 'speciesReference', {'species': sid, 'stoichiometry': str(stoic)})
            SubElement(machinery, 'capacity', self.machinery['capacity'].to_dict())
        if len(self.productions) + len(self.degradations) > 0:
            processings = SubElement(process, 'processings')
            if len(self.productions) > 0:
                productions = SubElement(processings, 'listOfProductions')
                attribs = {'processingMap': self.productions['processingMap'], 'set': self.productions['set']}
                processing = SubElement(productions, 'processing', attribs)
                inputs = SubElement(processing, 'listOfInputs')
                for sid in self.productions['inputs']:
                    SubElement(inputs, 'speciesReference', {'species': sid, 'stoichiometry': '1'})
            if len(self.degradations) > 0:
                degradations = SubElement(processings, 'listOfDegradations')
                attribs = {'processingMap': self.degradations['processingMap'], 'set': self.degradations['set']}
                processing = SubElement(degradations, 'processing', attribs)
                inputs = SubElement(processing, 'listOfInputs')
                for sid in self.degradations['inputs']:
                    SubElement(inputs, 'speciesReference', {'species': sid, 'stoichiometry': '1'})
        return process

    def to_dict(self):
        mach_capacity = ''
        mach_reactants = ''
        mach_products = ''
        if len(self.machinery) > 0:
            mach_capacity = ', '.join([f'{key}={val}'
                                       for key, val in self.machinery['capacity'].to_dict().items()])
            mach_reactants = '; '.join([f'species={species}, stoic={stoic}'
                                        for species, stoic in self.machinery['reactants'].items()])
            mach_products = '; '.join([f'species={species}, stoic={stoic}'
                                       for species, stoic in self.machinery['products'].items()])
        prod_pmap = ''
        prod_set = ''
        prod_inputs = ''
        if len(self.productions) > 0:
            prod_pmap = self.productions['processingMap']
            prod_set = self.productions['set']
            prod_inputs = ', '.join(self.productions['inputs'])

        degr_pmap = ''
        degr_set = ''
        degr_inputs = ''
        if len(self.degradations) > 0:
            degr_pmap = self.degradations['processingMap']
            degr_set = self.degradations['set']
            degr_inputs = ', '.join(self.degradations['inputs'])

        return {'process': self.id, 'name': self.name,
                'machineryCapacity': mach_capacity, 'machineryReactants': mach_reactants,
                'machineryProducts': mach_products,
                'productionProcessingMap': prod_pmap, 'productionSet': prod_set,
                'productionInputs': prod_inputs,
                'degradationProcessingMap': degr_pmap, 'degradationSet': degr_set,
                'degradationInputs': degr_inputs}


class RbaProcessingMap:

    def __init__(self, pmid, const_processing=None, comp_processing=None):
        self.id = pmid
        self.constant_processing = const_processing if type(const_processing) is dict else {}
        self.component_processings = comp_processing if type(comp_processing) is dict else {}

    @staticmethod
    def import_xml(processing_maps):

        data = {}
        for processing_map in processing_maps.findall('processingMap'):
            pmid = processing_map.attrib['id']
            rba_pmap = RbaProcessingMap(pmid)

            const_proc = processing_map.find('constantProcessing')
            if const_proc is not None:
                rba_pmap.constant_processing['reactants'] = get_species_refs_from_xml(
                    const_proc.find('listOfReactants'))
                rba_pmap.constant_processing['products'] = get_species_refs_from_xml(const_proc.find('listOfProducts'))

            comp_procs = processing_map.find('listOfComponentProcessings')
            if comp_procs is not None:
                for comp_proc in comp_procs.findall('componentProcessing'):
                    component = comp_proc.attrib['component']
                    cost = float(comp_proc.get('machineryCost', '0'))
                    reactants = get_species_refs_from_xml(comp_proc.find('listOfReactants'))
                    products = get_species_refs_from_xml(comp_proc.find('listOfProducts'))
                    rba_pmap.component_processings[component] = {'cost': cost, 'reactants': reactants,
                                                                 'products': products}
            data[pmid] = rba_pmap
        return data

    @staticmethod
    def from_df(df):
        data = {}
        for pmid, group in df.groupby('processingMap'):
            rba_pmap = RbaProcessingMap(pmid)
            for _, row in group.iterrows():
                component = row['component']
                reactants = get_species_refs_from_str(row['reactants'])
                products = get_species_refs_from_str(row['products'])
                if component == 'constantProcessing':
                    rba_pmap.constant_processing['reactants'] = reactants
                    rba_pmap.constant_processing['products'] = products
                else:
                    cost = row['machineryCost']
                    rba_pmap.component_processings[component] = {'cost': cost, 'reactants': reactants,
                                                                 'products': products}
            data[pmid] = rba_pmap
        return data

    def export_xml(self):
        pmap = Element('processingMap', {'id': self.id})

        if len(self.constant_processing) > 0:
            const_proc = SubElement(pmap, 'constantProcessing')
            if len(self.constant_processing['reactants']) > 0:
                reactants = SubElement(const_proc, 'listOfReactants')
                for sid, stoic in self.constant_processing['reactants'].items():
                    SubElement(reactants, 'speciesReference', {'species': sid, 'stoichiometry': str(stoic)})
            if len(self.constant_processing['products']) > 0:
                products = SubElement(const_proc, 'listOfProducts')
                for sid, stoic in self.constant_processing['products'].items():
                    SubElement(products, 'speciesReference', {'species': sid, 'stoichiometry': str(stoic)})

        if len(self.component_processings) > 0:
            comp_procs = SubElement(pmap, 'listOfComponentProcessings')
            for component, params in self.component_processings.items():
                attribs = {'component': component, 'machineryCost': str(params['cost'])}
                comp_proc = SubElement(comp_procs, 'componentProcessing', attribs)
                if len(params['reactants']) > 0:
                    reactants = SubElement(comp_proc, 'listOfReactants')
                    for sid, stoic in params['reactants'].items():
                        SubElement(reactants, 'speciesReference', {'species': sid, 'stoichiometry': str(stoic)})
                if len(params['products']) > 0:
                    products = SubElement(comp_proc, 'listOfProducts')
                    for sid, stoic in params['products'].items():
                        SubElement(products, 'speciesReference', {'species': sid, 'stoichiometry': str(stoic)})
        return pmap

    def to_df(self):
        pmid = self.id
        data = []
        if ((len(self.constant_processing.get('reactants', {})) > 0) or
                (len(self.constant_processing.get('reactants', {}))) > 0):
            cost = np.nan
            component = 'constantProcessing'
            reactants = '; '.join([f'species={species}, stoic={stoic}'
                                   for species, stoic in self.constant_processing['reactants'].items()])
            products = '; '.join([f'species={species}, stoic={stoic}'
                                  for species, stoic in self.constant_processing['products'].items()])
            data.append([pmid, component, cost, reactants, products])
        if len(self.component_processings) > 0:
            for component, comp_proc in self.component_processings.items():
                cost = comp_proc['cost']
                reactants = '; '.join([f'species={species}, stoic={stoic}'
                                       for species, stoic in comp_proc['reactants'].items()])
                products = '; '.join([f'species={species}, stoic={stoic}'
                                      for species, stoic in comp_proc['products'].items()])
                data.append([pmid, component, cost, reactants, products])
        df = pd.DataFrame(data, columns=['processingMap', 'component', 'cost', 'reactants', 'products'])
        return df

    def to_dict(self):
        const_proc = {}
        if 'reactants' in self.constant_processing:
            const_proc['reactants'] = '; '.join([f'species={species}, stoic={stoic}'
                                                 for species, stoic in self.constant_processing['reactants'].items()])
            const_proc['products'] = '; '.join([f'species={species}, stoic={stoic}'
                                                for species, stoic in self.constant_processing['products'].items()])

        comp_procs = {}
        if len(self.component_processings) > 0:
            for component, comp_proc in self.component_processings.items():
                mach_cost = comp_proc['cost']
                reactants = '; '.join([f'species={species}, stoic={stoic}'
                                       for species, stoic in comp_proc['reactants'].items()])
                products = '; '.join([f'species={species}, stoic={stoic}'
                                      for species, stoic in comp_proc['products'].items()])
                comp_procs[component] = {'machineryCost': mach_cost, 'reactants': reactants, 'products': products}

        return {'processing_map': self.id, 'constantProcessing': const_proc,
                'componentProcessings': comp_procs}



"""Implementation of FbcGeneProduct class.

Peter Schubert, HHU Duesseldorf, July 2022
"""
import re

from .sbml_sbase import SbmlSBase


class FbcGeneProduct(SbmlSBase):

    def __init__(self, s_gp):
        super().__init__(s_gp)
        self.label = s_gp['label']
        self.compartment = s_gp.get('compartment')

    @property
    def uid(self):
        uids = self.miriam_annotation.get_qualified_refs('bqbiol:is', 'uniprot')
        return uids[0] if len(uids) > 0 else f'GP_{self.label}'

    def modify_gene_label(self, pattern, replacement):
        if re.search(pattern, self.label):
            self.label = re.sub(pattern, replacement, self.label)

    def add_notes(self, notes):
        """Add gene id and name from Uniprot data.

        :param str notes: notes to add
        """
        self.notes = notes

    def modify_attribute(self, attribute, value):
        """Modify attribute value.

        :param str attribute: attribute name
        :param str value: value to be configured
        """
        setattr(self, attribute, value)

    def to_dict(self):
        data = super().to_dict()
        data['label'] = self.label
        return data



"""Implementation of Enzyme class.

Enzymes catalyze reactions

- an Enzyme can catalyze one or several reactions (promiscuity).
- a reaction can be catalzyed by different enzymes (isoenzyme).
- enzymes consist for gene products identified by uniprot ids
  at a given stoichiometry.
- enzyme composition need to be retrieved from e.g. Biocyc or .csv file
- based on composition we can determine a molecular weight.

froward and backward catalytic rates of an enzyme will be a reaction attribute.

Note for other types of models, e.g. RBA, enzymes could be assigned
 to a specific compartment or membrane.
 they could also be input to molecular machines like chaperoning
 or secretion apparatus

Here the implementation is primarily for enzyme constraint metabolic
models like GECKO, ccFBA, MOMENT

Peter Schubert, HHU Duesseldorf, June 2023
"""
import re

from ..utils.mapping_utils import get_srefs


class Enzyme:

    def __init__(self, enz_id, name, model, composition, compartment):
        """Instantiate enzyme object

        Enzyme compartment is set based on a reaction, catalyzed by the enzyme.
        Note: multiple compartments could be possible, only one is selected

        :param str enz_id: enzyme id, e.g. 'enz_b1094_b2836'
        :param str name: enzyme name
        :param model: reference to underlying model
        :type model: class:`XbaModel`
        :param composition: protein composition of enzyme
        :type composition: dict (key: locus/str, val: stoichiometry/float)
        :param str compartment: compartment of enzyme
        """
        self.id = enz_id
        self.name = name
        self.model = model
        self.composition = composition
        self.compartment = compartment
        self.active_sites = 1
        self.rids = []

    @property
    def mw(self):
        mw = 0.0
        for locus, stoic in self.composition.items():
            uid = self.model.locus2uid[locus]
            p = self.model.proteins[uid]
            mw += stoic * p.mw
        return mw

    @property
    def cofactors(self):
        cofactors = {}
        for locus, stoic in self.composition.items():
            uid = self.model.locus2uid[locus]
            p = self.model.proteins[uid]
            for cf_sid, cf_stoic in p.cofactors.items():
                if cf_sid not in cofactors:
                    cofactors[cf_sid] = 0.0
                cofactors[cf_sid] += stoic * cf_stoic
        return cofactors

    def add_reaction(self, rid):
        """Add reaction identifier.

        :param str rid: reaction id
        """
        self.rids.append(rid)

    def modify_attribute(self, attribute, value):
        """Modify attribute value.

        :param str attribute: attribute name
        :param value: value to be configured
        :type value: str or int or list[str]
        """
        if attribute == 'composition' and type(value) is str:
            value = get_srefs(re.sub('gene', 'species', value))
        setattr(self, attribute, value)



"""Implementation of TfaModel class.

Extend the XbaModel to become a thermodynamics constraint model.
The implementation of TFA is based on pyTFA (Salvy et al., 2019).
Thermodynamics data is converted to kJ/mol.

Peter Schubert, HHU Duesseldorf, October 2023
"""

import os
import re

import numpy as np
import pandas as pd
from collections import defaultdict
import pickle
import zlib

from .td_compartment_data import TdCompartmentData
from .td_species_data import TdSpeciesData
from .td_cue_data import TdCueData
from .td_reaction_data import TdReactionData
from .td_reactant_data import TdReactantData
from ..utils.tfa_utils import extract_atoms
from ..utils.mapping_utils import load_parameter_file
from ..xba_model.fbc_objective import FbcObjective
import f2xba.prefixes as pf


# TD parameters
GAS_CONSTANT = 8.314462/1000.0            # kJ mol-1 K-1
FARADAY_CONSTANT = 9.648533e4 / 1000.0    # kC mol-1  or kJ mol-1 V-1
CAL_PER_J = 4.184                         # cal/J

MAX_DRG = 1000.0                          # kJ/mol, maximum ∆rG' as variable bound
IRR_NO_TD_DRG0 = -200.0                   # kJ/mol, ∆rG'˚ below which an irr reactions requires no TD constraints
DEFAULT_DRG_ERROR = 2.0 * 4.184           # kJ/mol

# special metabolites protons and water
CPD_PROTON = 'cpd00067'        # seed id for protons (H) - not part of TD formulations
CPD_WATER = 'cpd00001'         # seed id for H2O - not used in reaction quotient


class TfaModel:
    """Extend a XbaModel to a thermodynamics constraint model.

    The implementation of TFA is based on Python package pyTFA (Salvy et al., 2019).

    .. code-block:: python

        xba_model = XbaModel('iJO1366.xml')
        xba_model.configure()

        tfa_model = TfaModel(xba_model)
        tfa_model.configure('iJO1366_tfa_parameters.xlsx')
        if tfa_model.validate():
            tfa_model.export('iJO1366_TFA.xml')
    """

    def __init__(self, xba_model):
        """Instantiate the TfaModel instance.

        :param xba_model: a reference to a XbaModel instance
        :type xba_model: :class:`f2xba.XbaModel`
        """
        self.model = xba_model
        self.td_params = {}
        self.td_compartments = {}
        self.td_species = {}
        self.td_cues = {}
        self.td_reactions = {}
        self.td_reactants = {}
        self.td_units = None
        self.temperature = 298.15 # temperature in Kevlin (25 ˚C)

    @property
    def rt(self):
        return GAS_CONSTANT * self.temperature

    def configure(self, fname):
        """Configuration with TFA configuration data.

        Accepted tables: 'general', 'td_compartments', 'modify_td_sids', 'modify_thermo_data',
        'modify_drg0_bounds'

        :param str fname: filename of TFA configuration file (.xlsx)
        :return: success
        :rtype: bool
        """
        sheet_names = ['general', 'td_compartments', 'modify_td_sids', 'modify_thermo_data', 'modify_drg0_bounds']
        tfa_params = load_parameter_file(fname, sheet_names)
        if 'general' not in tfa_params.keys():
            print(f'mandatory table "general" not found in the document')
            raise ValueError

        self.td_params = tfa_params['general']['value'].to_dict()
        self.temperature = 273.15 + self.td_params.get('temperature', 25.0)

        # load and configure thermodynamics data
        self._create_td_data(self.td_params['thermo_data_fname'], tfa_params)
        self._create_td_reactions()
        self._create_td_reactants()
        self._set_td_reactant_dfg0_tr()
        self._set_td_reaction_drg()

        # adding TFA related constraints and variables
        self._add_tfa_constraints()
        self._add_tfa_variables()

        # in case relaxed parameters are already available, implement them
        if 'modify_drg0_bounds' in tfa_params:
            dgr0_vids = []
            for vid in tfa_params['modify_drg0_bounds'].index:
                rid = re.sub(f'^{pf.V_DRG0_}', pf.R_, vid)
                if self.td_reactions[rid].add_td_constraints:
                    dgr0_vids.append(vid)
            self.modify_drg0_bounds(tfa_params['modify_drg0_bounds'].loc[dgr0_vids], remove_slack=False)

        # modify some model attributs and create L3V2 SBML model
        self.model.model_attrs['id'] += f'_TFA'
        if 'name' in self.model.model_attrs:
            self.model.model_attrs['name'] = f'TFA model of ' + self.model.model_attrs['name']
        self.model.sbml_container['level'] = 3
        self.model.sbml_container['version'] = 2

        self.print_td_stats()
        self.model.print_size()
        return True

    def export_slack_model(self, fname):
        """Create a slack model and export it as SBML coded file.

        The slack model is used for model parameter relaxation.

        :param str fname: filename (with extension '.xml' or '.xlsx')
        :return: success
        :rtype: bool
        """
        self._add_slack_variables()
        self._add_slack_objective()
        self.export(fname)

    def print_td_stats(self):
        """Print TD statistics on species/reactions
        """
        var_ids = set(self.model.reactions)
        sids = []
        td_sids = []
        mids = set()
        td_mids = set()
        for constr_id in self.model.species:
            if re.match('M_', constr_id):
                sidx = re.sub(f'^{pf.M_}', '', constr_id)
                midx = sidx.rsplit('_', 1)[0]
                sids.append(sidx)
                mids.add(midx)
                if pf.V_LC_ + sidx in var_ids:
                    td_sids.append(sidx)
                    td_mids.add(midx)

        orig_rids = set()
        td_rids = set()
        for var_id, r in self.model.reactions.items():
            if re.match(pf.R_, var_id):
                if r.kind in ['transporter', 'metabolic']:
                    ridx = re.sub(f'^{pf.R_}', '', r.orig_rid)
                    orig_rids.add(ridx)
                    if pf.V_DRG_ + ridx in var_ids:
                        td_rids.add(ridx)
        print(f'{len(td_sids)} species ({len(td_mids)} metabolites) with TD data from total {len(sids)} ({len(mids)})')
        print(f'{len(td_rids)} metabolic/transporter reactions with TD data from total {len(orig_rids)}')

    def integrate_min_slack(self, fba_fluxes):
        """Integrate optimization results from slack minimization of slack model.

        Called, with the flux values of a CobraPy optimiztion of the slack model

        Negative and positive slacks will be used to update bounds of related
        ∆Gr˚ variables.

        Summary of updates is returned

        :param fba_fluxes: CobraPy fluxes from solution object
        :type fba_fluxes: pandas.Series or dict (key: rid / str, val: flux / float)
        :return: summary of ∆Gr˚ bound updates performed
        :rtype: dict
        """
        drgo_relaxations = self._relax_drg0_variables(fba_fluxes)
        self._remove_slack_configuration()
        return drgo_relaxations

    def modify_drg0_bounds(self, df_modify_drg0_bounds, remove_slack=True):
        """Modify ∆rG'˚ bounds after model relaxation.

        We reused exising parameter ids for ∆rG'˚.
        These get updated with data from df_modify_drg0_bounds.
            index: variable id (str) for drg0 variables 'V_DRG0_<rid>'
            columns:
                component: set to 'reaction' / str
                attribute: either 'fbc_lower_bound' or 'fbc_upper_bound' / str
                value: new bound value / float

        :param df_modify_drg0_bounds: data on bounds to be updated
        :type df_modify_drg0_bounds: pandas.DataFrame
        :param remove_slack: if True, remove slack variables
        :type remove_slack: bool (default: True)
        """
        modify_attrs = {}
        for var_id, row in df_modify_drg0_bounds.iterrows():
            drg0_var = self.model.reactions[var_id]
            pid = getattr(drg0_var, row['attribute'])
            modify_attrs[pid] = ['parameter', 'value', row['value'], 'relaxation']
        cols = ['component', 'attribute', 'value', 'notes']
        df_modify_attrs = pd.DataFrame(modify_attrs.values(), index=list(modify_attrs), columns=cols)
        print(f"{len(df_modify_attrs):4d} ∆Gr'˚ variables need relaxation.")
        self.model.modify_attributes(df_modify_attrs, 'parameter')

        if remove_slack is True:
            self._remove_slack_configuration()

    def add_displacement(self):
        """Add displacement variable for ∆Gr.

        The subsequent investigation will explore how far reactions are operating from TD equilibrium.
        """
        self._add_displacement_constraints()
        self._add_displacement_variables()

    def validate(self):
        """Validate compliance to SBML standards.

        :return: success
        :rtype: bool
        """
        return self.model.validate()

    def export(self, fname):
        """Export TfaModel to SBML coded file or in tabular format.

        :param str fname: filename (with extension '.xml' or '.xlsx')
        :return: success
        :rtype: bool
        """
        return self.model.export(fname)

    def check_compatible_thermo_data(self, sid, metabolite_td_data):
        """Check compatibility between the model species corresponding thermo data record.

        For the check to pass, the TD record must have a valid ∆Gf° and be error free.
        Further, the formula and chemical charge, if configured on the model species, must
        be compatible with the formula and chemical charge configured in the TD record. The number
        of protons and corresponding charges are allowed to differ.

        :param str sid: species id in the model
        :param dict metabolite_td_data: metabolite record extracted from TD data
        :return: success/failure of test
        :rtype: bool
        """
        valid = False
        if metabolite_td_data['deltaGf_std'] < 1e6 and metabolite_td_data['error'] == 'Nil':
            s = self.model.species[sid]

            td_atom_delta = {}
            if hasattr(s, 'formula') and type(s.formula) is str:
                s_atoms = {atom: count for atom, count in extract_atoms(s.formula)}
                # Note: strings loaded via pickle from thermo db have type numpy.str_
                td_atoms = {atom: count for atom, count in extract_atoms(str(metabolite_td_data['formula']))}
                atoms = set(s_atoms.keys()).union(set(td_atoms.keys()))
                for atom in atoms:
                    delta = td_atoms.get(atom, 0) - s_atoms.get(atom, 0)
                    if delta != 0:
                        td_atom_delta[atom] = delta

            td_charge_delta = 999
            if hasattr(s, 'charge') and (isinstance(s.charge, float) or isinstance(s.charge, int)):
                td_charge_delta = metabolite_td_data['charge_std'] - s.charge

            if len(td_atom_delta) == 0 and td_charge_delta == 0:
                valid = True
            elif (len(td_atom_delta) == 1) and ('H' in td_atom_delta) and (td_atom_delta['H'] == td_charge_delta):
                valid = True
        return valid

    def _select_td_species(self, all_td_data, modify_td_ids):
        """Create TD species for species in the model.

        We make used of SEED refs in species model annotation and species name.
        We configure 'td_sid' attribute on model species to corresponding TD species,
          alternatively it is set to None
        only select TD species, that contain a valid formula

        Assignment of a td_sid is done in steps:
            1. we can set a specific td_sid in modify_td_sids
                - for this a metabolite id is extracted from the model species id using 'mid_regex_pattern'
            2. next, we try finding a matching metabolite name in thermo data
                - name matching is in lower case
            3. finaly we use one seed id from model annotation that exists in thermo data
                - first seed id in a sorted list of model seed references
                - Note: thermo data ids are based on seed ids, however they are not fully compliant to Model.Seed

            Finaly, we check that selected TD record has a valid ∆Gf° and is error free. In addition, the
            chemical formula and electrical charge, if configured on the model species, must be compatible
            with the TD data record. Only exception is a difference in protonation state and corresponding
            charge.

        If no valid TD record can be found fo a model species, reactions using the species cannot
        be configured with TD constraints.

        :param all_td_data: reference to all TD metabolite records in the TD database
        :type all_td_data: dict (key: td_sid / str, val: dict with TD data)
        :param modify_td_ids: mapping of metabolite ids to TD species id (seed compound id)
        :type modify_td_ids: dict (key: metabolite id / str, value: td_sid / str)
        :return: TD species ids that have been selected for the species in the model
        :rtype: set(str)
        """
        # for metabolite name mapping, we create a dict of names in thermo data to respective seed id
        name2td_sid = {}
        all_td_sids = set()
        for seed_id, td_sdata in all_td_data.items():
            if td_sdata['formula'] not in ['NA', 'NULL']:
                all_td_sids.add(seed_id)
                name2td_sid[td_sdata['name'].lower()] = seed_id
                for name in td_sdata['other_names']:
                    if len(name) > 0:
                        name2td_sid[name.lower()] = seed_id

        mid_regex_pattern = self.td_params.get('mid_regex_pattern', r'^M_(\w+)_\w+$')
        td_sids = set()
        for sid, s in self.model.species.items():
            selected_td_sid = None
            m = re.match(mid_regex_pattern, sid)
            if m:
                mid = m.group(1)
                if (mid in modify_td_ids) and (modify_td_ids[mid] in all_td_sids):
                    selected_td_sid = modify_td_ids[mid]
                else:
                    name = s.name.lower()
                    if name in name2td_sid:
                        selected_td_sid = name2td_sid[name]
                    else:
                        eligible_td_sids = set(s.seed_refs).intersection(all_td_sids)
                        if len(eligible_td_sids) > 0:
                            selected_td_sid = sorted(eligible_td_sids)[0]

                # check that formula and charges are compatible
                if selected_td_sid is not None:
                    if self.check_compatible_thermo_data(sid, all_td_data[selected_td_sid]) is False:
                        selected_td_sid = None
                s.modify_attribute('td_sid', selected_td_sid)
                if selected_td_sid is not None:
                    td_sids.add(selected_td_sid)

        return td_sids

    def _create_td_data(self, thermo_db_fname, tfa_params):
        """Load TD database and configure TD data based on TFA configuration file.

        We only extract records from the TD database that relate to our model.
        TD database must be formated as per TD database used by pyTFA (Salvy et al., 2019), see
        https://github.com/EPFL-LCSB/pytfa/raw/refs/heads/master/data/thermo_data.thermodb

        - Create td_compartments based on TFA configuration table 'td_compartments'.
        - Configure c_min and c_max on model species.
        - Load TD database from file (format must similar to pyTFA thermo_data.thermodb).
        - update TD database record attributes based on TFA configuration table 'modify_thermo_data'.
        - convert energy value from kcal/mol to kJ/mol, if units the TD database are in kcal/mol.
        - from TFA parameter file, table 'modify_td_sids', get fixed mapping from metabolite id to TD metabolite.
        - identify for each model species a valid TD record, based on TFA configuration table 'modify_td_sids',
          name matching and alternatively based on SEED compound ID configured. TD record must have valid
          ∆Gf°, be error free and formula/charge must be compatible. Protonation states can be different.
        - create TD species for species in the model where a valid TD record has been found

        :param str thermo_db_fname: file name of thermodynamics database
        :param tfa_params: configuration tables from TFA configuration file
        :type tfa_params: dict of pandas.DataFrames
        :return:
        """
        # TD configuration of compartments
        self.td_compartments = {cid: TdCompartmentData(str(cid), row)
                                for cid, row in tfa_params['td_compartments'].iterrows()}

        # set species concentrations, unless concentrations are configured already in XbaModel
        for sid, s in self.model.species.items():
            cid = s.compartment
            if hasattr(s, 'c_min') is False:
                s.modify_attribute('c_min', self.td_compartments[cid].c_min)
            if hasattr(s, 'c_max') is False:
                s.modify_attribute('c_max', self.td_compartments[cid].c_max)

        # load thermo dynamic data from file and correct some errors
        if os.path.exists(thermo_db_fname) is False:
            print(f'{thermo_db_fname} does not exist')
            raise FileNotFoundError
        with open(thermo_db_fname, 'rb') as file:
            all_thermo_data = pickle.loads(zlib.decompress(file.read()))

        # update TD data records based on TFA configuration table 'modify_thermo_data'
        if 'modify_thermo_data' in tfa_params:
            count = 0
            for td_sid, row in tfa_params['modify_thermo_data'].iterrows():
                if td_sid in all_thermo_data['metabolites']:
                    if row['attribute'] in all_thermo_data['metabolites'][td_sid]:
                        all_thermo_data['metabolites'][td_sid][row['attribute']] = row['value']
                        count += 1
                    else:
                        print(f'{td_sid} has no attribute {row["attribute"]} in TD database')
                else:
                    print(f'{td_sid} id not found in TD database (check table "modify_thermo_data")')
            print(f'{count:4d} thermo data attributes updated')

        modify_td_ids = {}
        if 'modify_td_sids' in tfa_params:
            modify_td_ids = tfa_params['modify_td_sids']['td_sid'].to_dict()

        # create td_metabolites with valid thermo data, filtered by required metabolites
        all_td_data = all_thermo_data['metabolites']
        td_sids = self._select_td_species(all_td_data, modify_td_ids)
        conv_factor = CAL_PER_J if all_thermo_data['units'] == 'kcal/mol' else 1.0
        self.td_species = {td_sid: TdSpeciesData(td_sid, all_td_data[td_sid], conv_factor) for td_sid in td_sids}
        n_td_sids = sum([1 for s in self.model.species.values() if s.td_sid is not None])
        not_supported = len(self.model.species) - n_td_sids
        print(f'{len(self.td_species):4d} metabolites with TD data covering {n_td_sids} model species; '
              f'({not_supported} not supported)')

        # create td_cues - selecting required cues only
        all_td_cues = all_thermo_data['cues']
        for td_sdata in self.td_species.values():
            to_del = set()
            for cue_id in td_sdata.struct_cues:
                if cue_id not in self.td_cues:
                    if cue_id in all_td_cues:
                        self.td_cues[cue_id] = TdCueData(cue_id, all_td_cues[cue_id], conv_factor)
                    else:
                        to_del.add(cue_id)
            # remove unsupported cues in td_metabolite, e.g. 'NO'
            if len(to_del) > 0:
                for cue_id in to_del:
                    del td_sdata.struct_cues[cue_id]

    def _create_td_reactions(self):
        """Create TD reactions for model reactions where TD data is fully available.

        TD reactions are created when all reactants and products have a valid TD data record with
        standard Gibbs free energy of formation.

        We remove transport reactions for water. Water can have different ∆fG'˚
        in different compartments and there would be no way to reverse reaction
        directionality, as water is not included in reaction quotient.

        Caluation of ∆fG'˚ is based on isomer group thermodynamics
        using least protonated species at the compartmental pH
        """
        not_supported = 0
        for rid, r in self.model.reactions.items():
            if r.kind in ['metabolic', 'transporter']:
                valid_dfg0_tr = True
                td_sids = set()
                for sid in list(r.reactants) + list(r.products):
                    if self.model.species[sid].td_sid is None:
                        valid_dfg0_tr = False
                        break
                    else:
                        td_sids.add(self.model.species[sid].td_sid)

                # do not consider reactions for pure water transport
                if td_sids == {CPD_WATER}:
                    valid_dfg0_tr = False

                if valid_dfg0_tr is True:
                    self.td_reactions[rid] = TdReactionData(rid, r.reversible, r.kind)
                else:
                    not_supported += 1
        print(f'{len(self.td_reactions):4d} reactions supported by TD data; ({not_supported} not supported)')

    def _create_td_reactants(self):
        """Create TD reactants for model species that have a valid TD data record.

        For each species in the model with TD data, create a TdReactant with same
        identifier and configure compartmental pH and ionic strength.
        """
        for sid, s in self.model.species.items():
            if s.td_sid:
                td_rdata = {'td_sid': s.td_sid,
                            'ph': self.td_compartments[s.compartment].ph,
                            'ionic_str': self.td_compartments[s.compartment].ionic_strength}
                self.td_reactants[sid] = TdReactantData(sid, td_rdata)

    def _set_td_reactant_dfg0_tr(self):
        """Calculate standard transformed Gibbs free energies of formation for TdReactants.

        Based on compartmental pH and ionic strength, calculate transformed Gibbs free energy
        of formation, determine avg number of protons and charge using on Isomer Group Theory.
        Update these values in TdReactant objects.
        """
        for sid, td_rdata in self.td_reactants.items():
            td_sdata = self.td_species[td_rdata.td_sid]
            dfg0_tr, avg_h_atoms_tr, avg_charge_tr = \
                td_sdata.get_std_transformed_gibbs_formation(td_rdata.ph, td_rdata.ionic_str, self.rt)
            td_rdata.modify_attribute('dfg0_tr', dfg0_tr)
            td_rdata.modify_attribute('avg_h_atoms_tr', avg_h_atoms_tr)
            td_rdata.modify_attribute('avg_charge_tr', avg_charge_tr)

    def _get_total_charge_transport(self, r):
        """Determine positive charge transport across membrane.

        Transfer of electrons, e.g. iJO1366 reaction R_FDH4pp, is implicitly supported.
        Charge transport across multiple compartments is supported.

        We use the charges configured for model species (i.e. biochemical reactants),
        as we assume that a specific charged chemical species from within the pseudoisomer group
        is transported and not a mixture of chemical species.

        We return total postive charge transfer from source to destination compartments.

        :param r: model reaction
        :type r: :class:`f2xba.xba_model.sbml_reaction.SbmlReaction`
        :return: direction of
        :rtype: dict (key: src_cid -> dest_cid / str, val: charge amount
        """
        charge_balance = {cid: 0.0 for cid in r.compartment.split('-')}
        if len(charge_balance) == 1:
            return {}

        for sid, stoic in r.reactants.items():
            s = self.model.species[sid]
            charge_balance[s.compartment] -= stoic * s.charge
        for sid, stoic in r.products.items():
            s = self.model.species[sid]
            charge_balance[s.compartment] += stoic * s.charge

        src = {cid: -charge for cid, charge in charge_balance.items() if charge < 0}
        dest = {cid: charge for cid, charge in charge_balance.items() if charge > 0}
        src_cids = list(src.keys())

        charge_transport = {}
        for dest_cid, charge in dest.items():
            for src_cid in src_cids:
                if src[src_cid] > 0.0:
                    if charge <= src[src_cid]:
                        charge_transport[f'{src_cid}->{dest_cid}'] = charge
                        src[src_cid] -= charge
                        break
                    else:
                        charge -= src[src_cid]
                        charge_transport[f'{src_cid}->{dest_cid}'] = src[src_cid]
                        src[src_cid] = 0.0
        return charge_transport

    def _get_std_transformed_gibbs_reaction(self, r):
        """Calculate standard transformed Gibbs energy of reaction

        both for reactions in single compartent and transport reactions

        Equation 4.4-2 in Alberty's book, 2003:
            ∆rG'˚ = ∑ v_i ∆fG'˚
            - ∆fG'˚: standard transformed Gibbs energy of formation for reactant at pH and ionic str

        in transformed properties protons do not appear in ∆fG'˚, see Alberty, 2003
            - ∆rG' = ∑^(N') v_i' µ_i'                                   (4.2-4)
            - ∆rG'˚ = ∑^(N') v_i' µ_i'˚                                 (4.2-7)
            - ∆G' = -S'dT + VdP + ∑^(Ns-1) µ_j' dn_j - RT ln(10) dpH    (4.1.13)
                - as µ_H+' = 0

        For transport reactions electrical work terms have to be added: F ∑ ∆φm_sd z_sd
                - ∆φm_sd: membrane potential (dest minus src potential) in V, e.g. -0.15 V for import p->c
                - z_sd: positive or negative charges transported along reaction direction (src -> dest)
                - electrical transported charges are summed up (based on model species charges)
                - charge transport spanning several compartments is considered

        :param r: model reaction
        :type r: :class:`f2xba.xba_model.sbml_reaction.SbmlReaction`
        :return: standard transformed Gibbs energy for reactions (or None for invalid)
        :rtype: float
        """
        srefs = {sid: -stoic for sid, stoic in r.reactants.items()}
        srefs |= {sid: stoic for sid, stoic in r.products.items()}

        # ∆rG'˚ = ∑ v_i ∆fG'˚
        drg0_tr = 0.0
        for sid, met_stoic in srefs.items():
            td_rdata = self.td_reactants[sid]
            # protons are removed for TD calculations when using transformed properties
            if td_rdata.td_sid != CPD_PROTON:
                drg0_tr += td_rdata.dfg0_tr * met_stoic

        # F ∑ ∆φm_sd z_sd
        charge_transport = self._get_total_charge_transport(r)
        if len(charge_transport) > 0:
            electrical_work = 0.0
            for transport_dir, charge in charge_transport.items():
                src_cid, dest_cid = transport_dir.split('->')
                membrane_pot = -self.td_compartments[dest_cid].membrane_pots[src_cid]
                electrical_work += FARADAY_CONSTANT * membrane_pot * charge
            drg0_tr += electrical_work

        return drg0_tr

    def _get_gibbs_reaction_error(self, r):
        """Calculate ∆Gr error related to ∆Gf of metabolites using group contribution method.

        The actual cues being modified by the reaction are considered for the error estimation.
        drg_error == sqrt(∑ (cue_est_error * stoic)^2)

        :param r: model reaction
        :type r: :class:`f2xba.xba_model.sbml_reaction.SbmlReaction`
        :return: estimated error relating to species ∆Gf (or 1e7 for invalid)
        :tupe: float
        """
        srefs = {sid: -stoic for sid, stoic in r.reactants.items()}
        srefs |= {sid: stoic for sid, stoic in r.products.items()}

        # for the reaction, we identify the actual cues being converted
        cues_balance = defaultdict(float)
        for sid, met_stoic in srefs.items():
            td_sid = self.td_reactants[sid].td_sid
            td_sdata = self.td_species[td_sid]
            for cue_id, cue_stoic in td_sdata.struct_cues.items():
                cues_balance[cue_id] += cue_stoic * met_stoic
        cues_unbalance = {cue_id: balance for cue_id, balance in cues_balance.items() if balance != 0.0}

        total_cues_dg_error = 0.0  # pyTFA calculation sqrt of sum of squared errors
        # TODO use sum of errors instead of sqrt of sum of squared errors
        # sum_cue_dfg_error_sum = 0.0  # alternative
        for cue_id, cues_stoic in cues_unbalance.items():
            cue_data = self.td_cues[cue_id]
            total_cues_dg_error += (cue_data.error * cues_stoic) ** 2
            # sum_cue_dfg_error_sum += cue_data.error * np.abs(cues_stoic)
        drg_error = np.sqrt(total_cues_dg_error)

        if drg_error == 0.0:
            drg_error = DEFAULT_DRG_ERROR

        return drg_error

    def _set_td_reaction_drg(self):
        """Configure TD properties ∆rG'˚ and error on td_reactions.

        Note: ∆rG'˚ are based on ∆fG'˚ and electrical work terms for transporters
        """
        for rid, td_rdata in self.td_reactions.items():
            r = self.model.reactions[rid]
            if td_rdata.kind in ['metabolic', 'transporter']:
                td_rdata.modify_attribute('drg0_tr', self._get_std_transformed_gibbs_reaction(r))
                td_rdata.modify_attribute('drg0_tr_error', self._get_gibbs_reaction_error(r))
                if td_rdata.reversible or td_rdata.drg0_tr > IRR_NO_TD_DRG0:
                    td_rdata.modify_attribute('add_td_constraints', True)
            else:
                print(f'{rid}, {td_rdata.kind} not supported for ∆Gr calculation')

    def _add_tfa_constraints(self):
        """Add TFA constraints to model as pseudio species.

        Note: irreversible reactions with TD data get not split
        and some constrations are not required
        """
        constr2name_rev = {pf.C_DRG_: "∆rG'",
                           pf.C_SU_: 'simultaneous use',
                           pf.C_GFC_: "∆rG' forward coupling",
                           pf.C_GRC_: "∆rG' reverse coupling",
                           pf.C_FFC_: 'flux forward coupling',
                           pf.C_FRC_: 'flux reverse coupling'}
        constr2name_irr = {pf.C_DRG_: "∆rG'",
                           pf.C_GFC_: "∆rG' forward coupling",
                           pf.C_FFC_: 'flux forward coupling'}

        pseudo_sids = {}
        for rid, td_rdata in self.td_reactions.items():
            if td_rdata.add_td_constraints:
                assert(td_rdata.drg0_tr is not None)
                ridx = re.sub(f'^{pf.R_}', '', rid)
                constr2name = constr2name_rev if td_rdata.reversible else constr2name_irr
                for prefix, name in constr2name.items():
                    pseudo_sids[prefix + ridx] = [f'{name} for {rid}', 'c', False, False, False]
        cols = ['name', 'compartment', 'hasOnlySubstanceUnits', 'boundaryCondition', 'constant']
        df_add_species = pd.DataFrame(pseudo_sids.values(), index=list(pseudo_sids), columns=cols)
        print(f'{len(df_add_species):4d} constraints to add')

        self.model.add_species(df_add_species)

    def _add_gibbs_reaction_variables(self):
        """Add ∆rG' and ∆rG'˚ variables to the model as pseudo reactions.

        based on pyTFA

        Variable V_DRG0_<rid> holds the value for the standard transformed Gibbs
        energy of reaction ∆rG'˚, which is constructed from standard transformed Gibbs
        energies of formation of participating reactants plus electrical work terms in case
        of charge transport across membranes.
        ∆rG'˚ is implemented as a variable that is allowed to vary between
        the computed value and the estimation error.

        Variable V_DRG_<rid> holds the value for the transformed Gibbs energy of reaction ∆rG',
        which is based on ∆rG'˚ and the reaction quotient, a summation over the
        ln metabolite concentrations.
        ∆rG' has 'unlimited' bounds.
        Constraint C_DRG_<rid> implements the calculation for ∆rG'
        - C_DRG_<rid>: V_DRG_<rid>  = V_DRG0_<rid> + RT ∑ v_i * V_LC_<sid>_i
        V_DRG_<rid> is  connected via coupling constraints C_GFC_<rid> and C_GRC_<rid>
        to forward V_FU_<rid> and reverse V_RU_<rid> use variables.
        """
        var2name = {'DRG': 'transformed Gibbs energy of reaction',
                    'DRG0': 'standard transformed Gibbs energy of reaction'}

        drg_lb_pid = self.model.get_fbc_bnd_pid(-MAX_DRG, 'kJ_per_mol', 'drg_lb', reuse=False)
        drg_ub_pid = self.model.get_fbc_bnd_pid(MAX_DRG, 'kJ_per_mol', 'drg_ub', reuse=False)

        pseudo_rids = {}
        for rid, td_rdata in self.td_reactions.items():
            if td_rdata.add_td_constraints:
                assert(td_rdata.drg0_tr is not None)
                ridx = re.sub(f'^{pf.R_}', '', rid)
                if td_rdata.reversible:
                    reactions_str = f'{pf.C_DRG_}{ridx} + {pf.C_GRC_}{ridx} -> {pf.C_GFC_}{ridx}'
                else:
                    reactions_str = f'{pf.C_DRG_}{ridx} -> {pf.C_GFC_}{ridx}'
                pseudo_rids[pf.V_DRG_ + ridx] = [f'{var2name["DRG"]} for {ridx}', reactions_str,
                                                 drg_lb_pid, drg_ub_pid, 'td_variable', 'continuous']

                drgo_lb = td_rdata.drg0_tr - td_rdata.drg0_tr_error
                drgo_ub = td_rdata.drg0_tr + td_rdata.drg0_tr_error
                drgo_lb_pid = self.model.get_fbc_bnd_pid(drgo_lb, 'kJ_per_mol', f'{pf.V_DRG0_}{ridx}_lb', reuse=False)
                drgo_ub_pid = self.model.get_fbc_bnd_pid(drgo_ub, 'kJ_per_mol', f'{pf.V_DRG0_}{ridx}_ub', reuse=False)
                pseudo_rids[pf.V_DRG0_ + ridx] = [f'{var2name["DRG0"]} for {ridx}', f'-> {pf.C_DRG_}{ridx}',
                                                  drgo_lb_pid, drgo_ub_pid, 'td_variable', 'continuous']

        cols = ['name', 'reactionString', 'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'notes']
        df_add_rids = pd.DataFrame(pseudo_rids.values(), index=list(pseudo_rids), columns=cols)
        print(f"{len(df_add_rids):4d} ∆rG'/∆rG'˚ variables to add")
        self.model.add_reactions(df_add_rids)

    def _add_use_variables(self):
        """Add forward and reverse use variables to the model as pseudo reactions.

        based on pyTFA

        Forward 'V_FU_<rid>' and Reverse 'V_RU_<rid>' use variables couple reaction
        directionality to ∆rG' via several constraints:
            - C_FFC_<rid>: connects forward reaction flux to V_FU_<rid>
                C_FFC_<rid>: R_<rid>_fwd - MAX_FLUX V_FU_<rid> <= 0
                i.e. if forward reaction flux > 0 -> V_FU_<rid> = 1
            - C_SU_<rid>: restricts simultaneous use of 'V_FU_<rid>' and 'V_BU_<rid>'
                C_SU_<rid>: V_FU_<rid> + V_RU_<rid> <= 1
                i.e. if V_FU_<rid> = 1 -> V_RU_<rid> = 0
            - C_FRC_<rid>: connects reverse reaction flux to V_RU_<rid>
                C_FRC_<rid>: R_<rid>_rev - MAX_FLUX V_RU_<rid> <= 0
                i.e. if V_RU_<rid> = 0 -> reverse reaction flux = 0
            - C_GFC_<rid>: couples negative ∆rG' (V_DRG_<rid>) to V_FU_<rid>
                C_GFC_<rid>: V_DRG_<rid> + MAX_DRG V_FU_<rid> <= MAX_DRG - eps
                eps = 1e-8
                i.e. if V_FU_<rid> = 1 -> V_DRG_<rid> < - eps
            - C_GRC_<rid>: couples positive ∆rG' (V_DRG_<rid>) to V_RU_<rid>
                C_GRC_<rid>: - V_DRG_<rid> + MAX_DRG V_RU_<rid> <= MAX_DRG - eps
                i.e. if V_RU_<rid> = 0 -> V_DRG_<rid> unconstraint
                i.e. if V_RU_<rid> = 1 -> V_DRG_<rid> >= eps

        Forward and backward use variables are binary variables (values 0 and 1) and must be
        configured as such in the linear problem before optimization,
        - e.g. in CobraPy: tfa_model.variables['V_FU_<rid>'].type = 'binary'
        Upper constraint bounds must be also configure in the linear problem.
        - e.g. tfa_slack_model.constraints['C_SU_<rid>'].lb = None
        - e.g. tfa_slack_model.constraints['C_SU_<rid>'].ub = 1.0
        """
        var2name = {'FU': 'forward use variable',
                    'RU': 'reverse use variable'}
        fbc_max_abs_flux_val = max(abs(self.model.fbc_flux_range[0]), abs(self.model.fbc_flux_range[1]))

        lb_pid = self.model.get_fbc_bnd_pid(0.0, 'fbc_dimensionless', 'binary_use_vars_lb', reuse=False)
        ub_pid = self.model.get_fbc_bnd_pid(1.0, 'fbc_dimensionless', 'binary_use_vars_ub', reuse=False)

        pseudo_rids = {}
        for rid, td_rdata in self.td_reactions.items():
            if td_rdata.add_td_constraints:
                assert(td_rdata.drg0_tr is not None)
                ridx = re.sub(f'^{pf.R_}', '', rid)
                if td_rdata.reversible:
                    fu_reaction = (f'{fbc_max_abs_flux_val} {pf.C_FFC_}{ridx} => '
                                   f'{MAX_DRG} {pf.C_GFC_}{ridx} + {pf.C_SU_}{ridx}')
                    bu_reaction = (f'{fbc_max_abs_flux_val} {pf.C_FRC_}{ridx} => '
                                   f'{MAX_DRG} {pf.C_GRC_}{ridx} + {pf.C_SU_}{ridx}')
                    pseudo_rids[pf.V_RU_ + ridx] = [f'{var2name["RU"]} for {ridx}', bu_reaction,
                                                    lb_pid, ub_pid, 'td_variable', 'binary']
                else:
                    fu_reaction = f'{fbc_max_abs_flux_val} {pf.C_FFC_}{ridx} => {MAX_DRG} {pf.C_GFC_}{ridx}'
                pseudo_rids[pf.V_FU_ + ridx] = [f'{var2name["FU"]} for {ridx}', fu_reaction,
                                                lb_pid, ub_pid, 'td_variable', 'binary']

        cols = ['name', 'reactionString', 'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'notes']
        df_add_rids = pd.DataFrame(pseudo_rids.values(), index=list(pseudo_rids), columns=cols)
        print(f'{len(df_add_rids):4d} forward/reverse use variables to add')
        self.model.add_reactions(df_add_rids)

    def _add_log_concentration_variables(self):
        """Add log concentration variables for reactants.

        log concentrations variables V_LC_<sid> hold the concentrations
        of reactants participating in reactions having complete TD data.

        log concentrations are required for calculation of Gibbs energy
        of reactions ∆rG' and are part of the reaction quotient
        The bound of the log concentrations are determined by min/max
        concentration values configured on species.

        Constraint C_DRG_<rid> contains the calculation for ∆rG'
        - C_DRG_<rid>: - V_DRG_<rid> + V_DRG0_<rid> + RT ∑ v_i * V_LC_<sid>_i = 0

        Protons (H) are not included in transformed Gibbs energy calculations, as
        proton concentration is hold constant by the compartment pH.
        Water (H2O) is generally not included in reaction quotient for reaction in diluted
        aqueous solutions.
        """
        var2name = {'LC': 'log reactant concentration'}

        # first we identify for each model species the reactions (including stoic) it participates in
        lc_variables = defaultdict(dict)
        for rid, td_rdata in self.td_reactions.items():
            if td_rdata.add_td_constraints:
                ridx = re.sub(f'^{pf.R_}', '', rid)
                r = self.model.reactions[rid]
                srefs = {sid: -stoic for sid, stoic in r.reactants.items()}
                srefs |= {sid: stoic for sid, stoic in r.products.items()}
                for sid, stoic in srefs.items():
                    td_rdata = self.td_reactants[sid]
                    # water and protons are not included in reaction quotient
                    if td_rdata.td_sid not in [CPD_PROTON, CPD_WATER]:
                        # if sid in lc_variables and ridx in lc_variables[sid]:
                        #     lc_variables[sid][ridx] += stoic * RT
                        # else:
                        lc_variables[sid][ridx] = stoic * self.rt

        # second we implement variables
        pseudo_rids = {}
        for lc_sid, data in lc_variables.items():
            sidx = re.sub(f'^{pf.M_}', '', lc_sid)
            c_min = self.model.species[lc_sid].c_min
            c_max = self.model.species[lc_sid].c_max
            lc_lb_pid = self.model.get_fbc_bnd_pid(np.log(c_min), 'kJ_per_mol', f'{pf.V_LC_}{sidx}_lb')
            lc_ub_pid = self.model.get_fbc_bnd_pid(np.log(c_max), 'kJ_per_mol', f'{pf.V_LC_}{sidx}_ub')
            reac_str = ' + '.join([f'{-rt_stoic} C_DRG_{ridx}' for ridx, rt_stoic in data.items() if rt_stoic < 0.0])
            prod_str = ' + '.join([f'{rt_stoic} C_DRG_{ridx}' for ridx, rt_stoic in data.items() if rt_stoic > 0.0])
            pseudo_rids[pf.V_LC_ + sidx] = [f'{var2name["LC"]} of {lc_sid}', f'{reac_str} -> {prod_str}',
                                            lc_lb_pid, lc_ub_pid, 'td_variable', 'continuous']

        cols = ['name', 'reactionString', 'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'notes']
        df_add_rids = pd.DataFrame(pseudo_rids.values(), index=list(pseudo_rids), columns=cols)
        print(f'{len(df_add_rids):4d} log concentration variables to add')
        self.model.add_reactions(df_add_rids)

    def _split_and_couple_reactions(self):
        """Split and couple reactions with TD information

        reactions with TD information get split into forward / reverse
        reactions are connected to forward/reverse coupling constraints
        """
        # fbc_max_flux_val = self.model.fbc_flux_range[1]
        # fbc_max_flux_pid = self.model.get_fbc_bnd_pid(fbc_max_flux_val, 'mmol_per_gDW', 'max_reverse_flux')

        count = 0
        count_opened_rev_dir = 0
        modify_attrs = []
        for rid, td_rdata in self.td_reactions.items():
            if td_rdata.add_td_constraints:
                ridx = re.sub(f'^{pf.R_}', '', rid)
                modify_attrs.append([rid, 'reaction', 'product', f'{pf.C_FFC_}{ridx}=1.0'])
                if td_rdata.reversible:
                    r = self.model.reactions[rid]
                    rev_r = self.model.split_reversible_reaction(r)
                    modify_attrs.append([rev_r.id, 'reaction', 'product', f'{pf.C_FRC_}{ridx}=1.0'])
                count += 1
        print(f'{count:4d} TD reactions split in forward/reverse, {count_opened_rev_dir} opened reverse direction')

        cols = ['id', 'component', 'attribute', 'value']
        df_modify_attrs = pd.DataFrame(modify_attrs, columns=cols)
        df_modify_attrs.set_index('id', inplace=True)
        print(f'{len(df_modify_attrs):4d} fwd/rev reactions to couple with flux direction')
        self.model.modify_attributes(df_modify_attrs, 'reaction')

    def _add_rhs_variables(self):
        """Add (fixed) RHS variables so RHS can be left at zero:
        """
        one_mmol_pid = self.model.get_fbc_bnd_pid(1.0, self.model.flux_uid, 'one_mmol', reuse=False)
        one_kj_pid = self.model.get_fbc_bnd_pid(1.0, 'kJ_per_mol', 'one_kJ', reuse=False)

        rhs_vars = {}
        rhs_flux_reactants = {sid: 1.0 for sid in self.model.species if re.match(pf.C_SU_, sid)}
        rhs_energy_reactants = {sid: 999.99 for sid in self.model.species
                                if re.match(pf.C_GFC_, sid) or re.match(pf.C_GRC_, sid)}
        rhs_vars[pf.V_RHS_FC] = [f'RHS for TD C_SU_ constraints', rhs_flux_reactants, {}, False,
                                 one_mmol_pid, one_mmol_pid, 'fixed variable', 'continuous']
        rhs_vars[pf.V_RHS_GC] = [f'RHS for TD C_GxC_ constraints', rhs_energy_reactants, {}, False,
                                 one_kj_pid, one_kj_pid, 'fixed variable', 'continuous']
        cols = ['name', 'reactants', 'products', 'reversible', 'fbcLowerFluxBound', 'fbcUpperFluxBound',
                'kind', 'notes']
        df_add_rids = pd.DataFrame(rhs_vars.values(), index=list(rhs_vars), columns=cols)
        print(f"{len(rhs_vars):4d} RHS variables to add")
        self.model.add_reactions(df_add_rids)

    def _add_tfa_variables(self):
        """Add TFA related variables to the model as pseudo reactions.

        based on pyTFA
        TFA specific variables are created and connected to constraints. Variables include:
        - [standard] transformed Gibbs energy of reaction variables ∆rG'˚ and ∆rG'
        - forward/reverse use binary variables for flux direction / ∆rG' coupling
        - log concentration variables for reactants
        - split reactions with TD info in fwd/rev and connect to coupling constraints
        - right-hand-side variables with fixed values to handle RHS not equal zero
        """
        self._add_gibbs_reaction_variables()
        self._add_use_variables()
        self._add_log_concentration_variables()
        self._split_and_couple_reactions()
        self._add_rhs_variables()

        print(f'{len(self.model.parameters):4d} parameters')

    # SLACK MODEL CONFIGURATION / TFA MODEL RELAXATION
    def _add_slack_variables(self):
        """Add slack variables for ∆rG'˚ relaxation.

        Negative and positive Slack variables are introduced according
        to pyTFA relaxation_drgo() method.
        Negative slack variables consume ∆rG' constraint (C_DRG_<rid>),
        positive slack variables produce ∆rG' constraint.
        Both slack variables are configured as non-negative (lower bound = 0)

        Following FBA optimization of the slack model (minimize sum of slack variables),
        the values of the slack variables can be used to adjust related ∆rG'˚ variable bounds.

        Slack objective and slack variables need to be removed for final TFA model
        """
        var2name = {'NS': "negative slack variable on ∆rG'˚",
                    'PS': "positive slack variable on ∆rG˚'"}

        lb_pid = self.model.get_fbc_bnd_pid(0.0, 'kJ_per_mol', 'slack_lb', reuse=False)
        ub_pid = self.model.get_fbc_bnd_pid(MAX_DRG, 'kJ_per_mol', 'slack_ub', reuse=False)

        pseudo_rids = {}
        for rid, td_rdata in self.td_reactions.items():
            if td_rdata.add_td_constraints:
                assert(td_rdata.drg0_tr is not None)
                ridx = re.sub(f'^{pf.R_}', '', rid)
                pseudo_rids[pf.V_NS_ + ridx] = [f'{var2name["NS"]} for {ridx}', f'{pf.C_DRG_}{ridx} =>',
                                                lb_pid, ub_pid, 'td_variable', 'continuous']
                pseudo_rids[pf.V_PS_ + ridx] = [f'{var2name["PS"]} for {ridx}', f'=> {pf.C_DRG_}{ridx}',
                                                lb_pid, ub_pid, 'td_variable', 'continuous']

        cols = ['name', 'reactionString', 'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'notes']
        df_add_rids = pd.DataFrame(pseudo_rids.values(), index=list(pseudo_rids), columns=cols)
        print(f"{len(df_add_rids):4d} slack variables for ∆rG'˙ to add")
        self.model.add_reactions(df_add_rids)

    def _add_slack_objective(self):
        """Add the slack objective (minimize sum over slack variables).

        Active optimization objective of model is inactivated
        The new slack objective is created and added to the model.
        """
        # inactivate active optimization objective
        for obj_id, obj in self.model.objectives.items():
            if obj.active is True:
                self.td_params['objective_id'] = obj_id
                obj.modify_attribute('active', False)

        # create new active slack_objective (minimize sum of slack variables)
        slack_obj_id = 'slack_objective'
        slack_obj_coefs = [rid for rid in self.model.reactions
                           if re.match(pf.V_PS_, rid) or re.match(pf.V_NS_, rid)]
        srefs_str = '; '.join([f'reac={var_name}, coef=1.0' for var_name in slack_obj_coefs])
        slack_obj_dict = {'type': 'minimize', 'active': True, 'fluxObjectives': srefs_str}
        self.model.objectives[slack_obj_id] = FbcObjective(pd.Series(slack_obj_dict, name=slack_obj_id))

    def _relax_drg0_variables(self, fluxes):
        """Relax ∆rG'˚ variable bounds based on slack minimization of slack model.

        Nonzero negative / positive slack variables are identified and
        bounds of related ∆rG'˚ variables are updated.
        E.g. upper bound update for non-zero positive slack variable is:
            - old upper bound value + slack + epsilon (1e-6)
        Note: we are updating the values of already configured parameter ids.

        A summary of updated values is returned

        :param fluxes: CobraPy fluxes from solution object
        :type fluxes: pandas.Series or dict (key: rid / str, val: flux / float)
        :return: summary of ∆Gr˚ bound updates performed
        :rtype: dict
        """
        eps = 0.5
        drg0_slack = {}
        for vname, flux in fluxes.items():
            if flux > 0.0 and (re.match(pf.V_NS_, vname) or re.match(pf.V_PS_, vname)):
                drg0_slack[vname] = flux

        drgo_relaxations = {}
        modify_attrs = {}
        for vname, slack in drg0_slack.items():
            if re.match(pf.V_PS_, vname):
                ridx = re.sub(f'^{pf.V_PS_}', '', vname)
            else:
                ridx = re.sub(f'^{pf.V_NS_}', '', vname)
            drg0_var = self.model.reactions[pf.V_DRG0_ + ridx]
            drg0_lb_pid = drg0_var.fbc_lower_bound
            drg0_ub_pid = drg0_var.fbc_upper_bound
            drg0_lb_val = self.model.parameters[drg0_lb_pid].value
            drg0_ub_val = self.model.parameters[drg0_ub_pid].value

            if re.match(pf.V_NS_, vname):
                drg0_new_lb_val = drg0_lb_val - slack - eps
                modify_attrs[drg0_lb_pid] = ['parameter', 'value', drg0_new_lb_val, 'relaxation']
                drgo_relaxations[ridx] = f"lower ∆rG'˚ bound from {drg0_lb_val:8.4f} to {drg0_new_lb_val:8.4f} " \
                                         f'[{drg0_new_lb_val:8.4f}, {drg0_ub_val:8.4f}]'
            else:
                drg0_new_ub_val = drg0_ub_val + slack + eps
                modify_attrs[drg0_ub_pid] = ['parameter', 'value', drg0_new_ub_val, 'relaxation']
                drgo_relaxations[ridx] = f"upper ∆rG'˚ bound from {drg0_ub_val:8.4f} to {drg0_new_ub_val:8.4f} " \
                                         f'[{drg0_lb_val:8.4f}, {drg0_new_ub_val:8.4f}]'

        df_modify_attrs = pd.DataFrame(modify_attrs, index=['component', 'attribute', 'value', 'notes']).T
        print(f"{len(df_modify_attrs):4d} ∆Gr'˚ variables need relaxation.")
        self.model.modify_attributes(df_modify_attrs, 'parameter')
        return drgo_relaxations

    def _remove_slack_configuration(self):
        """Remove the slack configuration (variables and objective)

        Remove slack objective and negative/positive slack variables.
        Reactivate the old optimization objective
        """
        del self.model.objectives['slack_objective']
        to_del = [vname for vname in self.model.reactions
                  if re.match(pf.V_PS_, vname) or re.match(pf.V_NS_, vname)]
        for vname in to_del:
            del self.model.reactions[vname]

        old_obj_id = self.td_params['objective_id']
        self.model.objectives[old_obj_id].modify_attribute('active', True)

    # ADDING DISPLACEMENT VARIABLES AND CONSTRAINTS
    def _add_displacement_constraints(self):
        """Add displacement constraints wrt to ∆Gr.

        based on pyTFA

        Adding new species (constraints) to the model
        """
        constr2name = {'C_DC': "∆rG' displacement constraint"}

        pseudo_sids = {}
        for rid, td_rdata in self.td_reactions.items():
            if td_rdata.add_td_constraints:
                ridx = re.sub(f'^{pf.R_}', '', rid)
                for constr, name in constr2name.items():
                    pseudo_sids[f'{constr}_{ridx}'] = [f'{name} for {rid}', 'c', False, False, False]

        cols = ['name', 'compartment', 'hasOnlySubstanceUnits', 'boundaryCondition', 'constant']
        df_add_species = pd.DataFrame(pseudo_sids.values(), index=list(pseudo_sids), columns=cols)
        print(f"{len(df_add_species):4d} ∆rG' displacements constraints to add")
        self.model.add_species(df_add_species)

    def _add_displacement_variables(self):
        """Add variables for thermodynamic displacement.

        based on pyTFA, see Salvy et al. 2018

        Thermodynamic displacement variables V_LNG_<rid> hold information how far
        the transformed Gibbs energy of reaction is away from equilibrium.
            V_LNG_<rid> = V_DRG_<rid> / RT
        exp(V_LNG_<rid>) give a factor that multiplied with a product concentration
        results in equilibrium.

        The equation is implemented by a displacement constraint C_DC_<rid>
            C_DC_<rid>: - V_DRG_<rid> / RT + V_LNG_<rid> = 0
        """
        var2name = {'LNG': 'thermodynamic displacement'}

        lb_pid = self.model.get_fbc_bnd_pid(-MAX_DRG, 'fbc_dimensionless', 'displacement_lb', reuse=False)
        ub_pid = self.model.get_fbc_bnd_pid(MAX_DRG, 'fbc_dimensionless', 'displacement_ub', reuse=False)

        pseudo_rids = {}
        for rid, td_rdata in self.td_reactions.items():
            if td_rdata.add_td_constraints:
                ridx = re.sub(f'^{pf.R_}', '', rid)
                pseudo_rids[pf.V_LNG_ + ridx] = [f'{var2name["LNG"]} for {ridx}', f'-> {pf.C_DC_}{ridx}',
                                                 lb_pid, ub_pid, 'td_variable', 'continuous']

        cols = ['name', 'reactionString', 'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'notes']
        df_add_rids = pd.DataFrame(pseudo_rids.values(), index=list(pseudo_rids), columns=cols)
        print(f'{len(df_add_rids):4d} thermodynamic displacement variables to add')
        self.model.add_reactions(df_add_rids)

        # connect DG variables to displacement constraints
        modify_attrs = {}
        for rid, td_rdata in self.td_reactions.items():
            ridx = re.sub(f'^{pf.R_}', '', rid)
            modify_attrs[pf.V_DRG_ + ridx] = ['reaction', 'reactant', f'{pf.C_DC_}{ridx}={1.0/self.rt}']

        cols = ['component', 'attribute', 'value']
        df_modify_attrs = pd.DataFrame(modify_attrs.values(), index=list(modify_attrs), columns=cols)
        print(f"{len(df_modify_attrs):4d} ∆rG' variables coupled to thermodynamic displacement constraints")
        self.model.modify_attributes(df_modify_attrs, 'reaction')



"""Implementation of RbaMacromolecules, RbaComponent and RbaMacromolecule classes.

Peter Schubert, CCB, HHU Duesseldorf, December 2022
"""

import os
import numpy as np
import pandas as pd
from xml.etree.ElementTree import parse, ElementTree, Element, SubElement, indent
from ..utils.calc_mw import get_seq_composition

type2tag = {'dna': 'RBADna', 'rna': 'RBARnas', 'protein': 'RBAProteins'}


class RbaMacromolecules:

    def __init__(self, mm_type):
        self.type = mm_type
        self.components = {}
        self.macromolecules = {}

    def import_xml(self, model_dir):

        file_name = os.path.join(model_dir, self.type + '.xml')
        if os.path.exists(file_name) is True:
            root = parse(file_name).getroot()
            assert root.tag == type2tag[self.type]
            self.components = RbaComponent.import_xml(root.find('listOfComponents'))
            self.macromolecules = RbaMacromolecule.import_xml(root.find('listOfMacromolecules'))
        else:
            print(f'{file_name} not found!')

    def from_df(self, m_dict):
        if self.type in m_dict:
            self.components = RbaComponent.from_df(m_dict[self.type])
            self.macromolecules = RbaMacromolecule.from_df(m_dict[self.type])
        else:
            print(f'{self.type} not imported!')

    def add_macromolecule(self, mm_id, c_name, composition):
        """Add a macromolecule and determine weight.

        :param str mm_id: id of macromolecule
        :param str c_name: RBA compartment name
        :param composition: composition
        :type composition: dict (key: component/str, val: stoic/float)
        :return:
        """
        if mm_id not in self.macromolecules:
            mm = RbaMacromolecule(mm_id, compartment=c_name, composition=composition)
            for comp_id, stoic in mm.composition.items():
                mm.weight += stoic * self.components[comp_id].weight
            self.macromolecules[mm_id] = mm

    def from_xba(self, rba_params, xba_model, cid_mappings):
        """Configure macromolecules.

        Components get configured based on RBA parameters
        - for proteins, also cofactors used in proteins are added as components

        Macromolecules get configured based xba model and RBA parameters
        - 'dna': 'dna' is added with relative nucleotide composition based on DNA GC content
        - 'rna': 'mrna' is added with relative nucleotide compostion based on all mRNAs in DNA
            - tRNAs and rRNAs are added from RBA parameters using NCBI derived composition
        - 'protein': add proteins from XBA model

        :param rba_params: RBA model specific parametrization
        :type rba_params: dict of pandas DataFrames
        :param xba_model: xba model based on genome scale metabolic model
        :type xba_model: class:`XbaModel`
        :param dict cid_mappings: dictionary of specific compartment names
        """
        df_pmaps_data = rba_params['processing_maps']
        df_mach_data = rba_params['machineries']
        df_trna_data = rba_params['trna2locus']
        df_components = df_pmaps_data[df_pmaps_data['set'] == self.type]
        cytoplasm_cid = cid_mappings['cytoplasm_cid']
        rcid2cid = cid_mappings['rcid2cid']

        # configure components
        if self.type in ['dna', 'rna']:
            for _, row in df_components.iterrows():
                cmp_id = row['component']
                name = row['name'] if type(row['name']) is str else None
                self.components[cmp_id] = RbaComponent(cmp_id, name=name, c_type='nucleotide', weight=row['weight'])

        if self.type == 'protein':
            for _, row in df_components.iterrows():
                cmp_id = row['component']
                if cmp_id != 'cofactor':
                    name = row['name'] if type(row['name']) is str else None
                    self.components[cmp_id] = RbaComponent(cmp_id, name=name, c_type='amino_acid', weight=row['weight'])
            # TODO: move cofactor to enzyme level for SBML model
            for p in xba_model.proteins.values():
                for sid in p.cofactors:
                    if sid not in self.components:
                        name = xba_model.species[sid].name
                        self.components[sid] = RbaComponent(sid, name=name, c_type='cofactor', weight=0.0)

        # configure macromolecules
        if self.type == 'dna':
            # add DNA composition using average DNA composition
            gc = xba_model.ncbi_data.get_gc_content()
            rel_conc = {'G': gc / 2.0, 'C': gc / 2.0, 'A': (1.0 - gc) / 2.0, 'T': (1.0 - gc) / 2.0}
            self.macromolecules['dna'] = RbaMacromolecule('dna', compartment=cytoplasm_cid, composition=rel_conc)

        if self.type == 'rna':
            # add mRNA using average mRNA composition
            mrna_avg_comp = xba_model.ncbi_data.get_mrna_avg_composition()
            self.macromolecules['mrna'] = RbaMacromolecule('mrna', compartment=cytoplasm_cid,
                                                           composition=mrna_avg_comp)
            # add tRNAs
            for trna_id, row in df_trna_data.iterrows():
                cid = rcid2cid[row['compartment']]
                record = xba_model.ncbi_data.locus2record[row['label']]
                composition = record.spliced_nt_composition
                self.macromolecules[trna_id] = RbaMacromolecule(trna_id, compartment=cid, composition=composition)
            # add rRNAs
            for _pm, row in df_mach_data[df_mach_data['set'] == 'rna'].iterrows():
                cid = rcid2cid[row['compartment']]
                record = xba_model.ncbi_data.locus2record[row['label']]
                rrna_id = row['id']
                composition = record.spliced_nt_composition
                self.macromolecules[rrna_id] = RbaMacromolecule(rrna_id, compartment=cid, composition=composition)
        if self.type == 'protein':
            for p in xba_model.proteins.values():
                locus = p.locus
                cid = rcid2cid[p.compartment]
                composition = get_seq_composition(p.aa_sequence) | p.cofactors
                self.macromolecules[locus] = RbaMacromolecule(locus, compartment=cid, composition=composition)

        # calulate macromolecular weight
        for mm in self.macromolecules.values():
            for comp_id, stoic in mm.composition.items():
                mm.weight += stoic * self.components[comp_id].weight

        print(f'{len(self.macromolecules):4d} macromolecules ({len(self.components)} components) in {self.type} ')

    def export_xml(self, model_dir):

        mm_set = {'rna': 'rnas', 'protein': 'proteins'}.get(self.type, self.type)
        file_name = os.path.join(model_dir, mm_set + '.xml')
        root = Element(type2tag[self.type])

        components = SubElement(root, 'listOfComponents')
        for item in self.components.values():
            components.append(item.export_xml())

        macromolecules = SubElement(root, 'listOfMacromolecules')
        for item in self.macromolecules.values():
            macromolecules.append(item.export_xml())

        tree = ElementTree(root)
        indent(tree)
        tree.write(file_name)

    def to_df(self):

        if len(self.components) > 0:
            df_comp = pd.DataFrame([item.to_dict() for item in self.components.values()])
        else:
            df_comp = pd.DataFrame(columns=['component', 'name', 'type', 'weight'])
        df_comp.set_index('component', inplace=True)

        if len(self.macromolecules) > 0:
            df_mm = pd.DataFrame([item.to_dict() for item in self.macromolecules.values()])
        else:
            df_mm = pd.DataFrame(columns=['id', 'compartment'])
        df_mm.set_index('id', inplace=True)
        cols = ['compartment'] + df_comp.index.to_list()
        df = pd.concat((df_comp.T, df_mm)).reindex(columns=cols)

        mm_set = {'rna': 'rnas', 'protein': 'proteins'}.get(self.type, self.type)
        df.index.name = mm_set
        return {mm_set: df}


class RbaComponent:

    def __init__(self, cid, name=None, c_type=None, weight=np.nan):
        self.id = cid
        self.name = name
        self.type = c_type
        self.weight = weight

    @staticmethod
    def import_xml(components):
        data = {}
        for component in components.findall('component'):
            cid = component.attrib['id']
            rba_component = RbaComponent(cid)
            rba_component.name = component.get('name')
            rba_component.type = component.get('type')
            rba_component.weight = float(component.attrib['weight'])
            data[cid] = rba_component
        return data

    @staticmethod
    def from_df(df):
        data = {}
        components = [col for col in df.columns if col != 'compartment']
        for cid in components:
            rba_component = RbaComponent(cid)
            if type(df.at['name', cid]) is str:
                rba_component.name = df.at['name', cid]
            if type(df.at['type', cid]) is str:
                rba_component.type = df.at['type', cid]
            rba_component.weight = df.at['weight', cid]
            data[cid] = rba_component
        return data

    def export_xml(self):
        attribs = {'id': self.id}
        if type(self.name) is str:
            attribs |= {'name': self.name}
        if type(self.type) is str:
            attribs |= {'type': self.type}
        attribs |= {'weight': str(self.weight)}
        return Element('component', attribs)

    def to_dict(self):
        return {'component': self.id, 'name': self.name,
                'type': self.type, 'weight': self.weight}


class RbaMacromolecule:

    def __init__(self, mmid, compartment='', composition=None):
        self.id = mmid
        self.compartment = compartment
        self.composition = composition if type(composition) is dict else {}
        self.weight = 0.0
        self.sid = None
        self.scale = 1000.0

    @staticmethod
    def import_xml(macromolecules):

        data = {}
        for macromolecule in macromolecules.findall('macromolecule'):
            mmid = macromolecule.attrib['id']
            rba_macromolecule = RbaMacromolecule(mmid)
            rba_macromolecule.compartment = macromolecule.attrib['compartment']
            composition = macromolecule.find('composition')
            for component_ref in composition.findall('componentReference'):
                component = component_ref.attrib['component']
                stoic = float(component_ref.attrib['stoichiometry'])
                rba_macromolecule.composition[component] = stoic
            data[mmid] = rba_macromolecule
        return data

    @staticmethod
    def from_df(df):
        data = {}
        components = [col for col in df.columns if col != 'compartment']
        macromolecules = [idx for idx in df.index if idx not in ['name', 'type', 'weight']]
        for mmid in macromolecules:
            rba_macromolecule = RbaMacromolecule(mmid)
            rba_macromolecule.compartment = df.at[mmid, 'compartment']
            for component in components:
                stoic = df.at[mmid, component]
                if np.isfinite(stoic) and stoic > 0.0:
                    rba_macromolecule.composition[component] = stoic
            data[mmid] = rba_macromolecule

        return data

    def export_xml(self):
        attribs = {'id': self.id, 'compartment': self.compartment}
        macromolecule = Element('macromolecule', attribs)
        composition = SubElement(macromolecule, 'composition')
        for component, stoic in self.composition.items():
            SubElement(composition, 'componentReference', {'component': component, 'stoichiometry': str(stoic)})
        return macromolecule

    def to_dict(self):
        return {'id': self.id, 'compartment': self.compartment} | self.composition



"""Implementation of TdCompartmentData class.

holding thermodynamics data for a compartment

Peter Schubert, HHU Duesseldorf, Octobert 2023
"""
import re


class TdCompartmentData:

    def __init__(self, cid, c_data):
        """Instantiate thermodynamics compartment data.

        Collect information supplied
        convert membrane potentials from mV to V

        c_data expected to have the keys:
          'ph': compartment pH
          'ionic_strenght_M': ionic strength in mol/l
          'c_min_M': minimum metabolite concentrations in mol/l
          'c_max_M': maximum metabolite concentrations in mol/l
          '<cid>_mV': membrane potential in mV (other compartment potential - own compartment potential)

        :param str cid: compartment id
        :param c_data: thermodynamics data
        :type c_data: dict or dict-like (e.g. pandas.Series)
        """
        self.id = cid
        self.ph = c_data['ph']
        self.ionic_strength = c_data['ionic_strength_M']
        self.c_min = c_data['c_min_M']
        self.c_max = c_data['c_max_M']
        self.membrane_pots = {}
        for key, val in c_data.items():
            if key.endswith('_mV'):
                other_cid = re.sub('_mV$', '', key)
                self.membrane_pots[other_cid] = val / 1000.0   # V -> mV



"""Implementation of NcbiFtRecord class.

Peter Schubert, CCB, HHU Duesseldorf, January 2024
"""
import re
from collections import defaultdict


list_attrs = {'db_xref', 'EC_number'}
map_complement = {'A': 'T', 'T': 'A', 'G': 'C', 'C': 'G'}


class NcbiFtRecord:

    def __init__(self, record_type, start, stop):
        self.type = record_type
        self.regions = [[start, stop]]
        self.forward = True if stop > start else False

    def add_region(self, start, stop):
        self.regions.append([start, stop])

    def add_attribute(self, attr_id, attr_val):
        if attr_id not in list_attrs:
            setattr(self, attr_id, attr_val)
        else:
            if hasattr(self, attr_id):
                getattr(self, attr_id).append(attr_val)
            else:
                setattr(self, attr_id, [attr_val])

    @staticmethod
    def sequence_reverse(seq):
        """Reverse a DNA sequence.

        :param seq: sequence of DNA nucleotides
        :type seq: str consisting of letters ATGC
        :return: complement of nucleotide sequence
        :rtype: str
        """
        compl = ''
        for nt in seq[::-1]:
            compl += map_complement[nt]
        return compl

    def get_composition(self, chrom_seq):
        composition = defaultdict(int)
        for (start, stop) in self.regions:
            if self.forward is True:
                dna_sequence = chrom_seq[start - 1:stop]
            else:
                dna_sequence = self.sequence_reverse(chrom_seq[stop - 1:start])
            rna_sequence = re.sub('T', 'U', dna_sequence)
            nucleotides = sorted(set(rna_sequence))
            for nt in nucleotides:
                composition[nt] += rna_sequence.count(nt)
        return dict(composition)

    @property
    def length(self):
        length = 0
        for (start, stop) in self.regions:
            length += abs(stop - start) + 1
        return length

    @property
    def old_locus(self):
        return getattr(self, 'old_locus_tag', None)

    @property
    def locus(self):
        return getattr(self, 'locus_tag', None)



"""Implementation of CobraRbaResults class.

Process results obtained by CobraPy optimization of SBML coded RBA models
accross several conditions.

This works on Cobra Model and considers additional constraints
and variables introduced by RBA formalism.

Peter Schubert, HHU Duesseldorf, December2023
"""

import re
import pandas as pd
from collections import defaultdict

import f2xba.prefixes as pf
from .results import Results


class RbaResults(Results):
    """Optimization results processing for RBA models generated by f2xba.

    Processing and presentation of optimization results for a single or several
    (media) conditions.
    """


    def __init__(self, optim, results, df_mpmf=None):
        """Instantiate the RbaResults instance.

        :param optim: RbaOptimization instance
        :type optim: :class:`RbaOptimization`
        :param results: conditions with optimization results
        :type results: dict (key: str, val: :class:`Solution`)
        :param df_mpmf: measured protein mass fractions for conditions (default: None)
        :type df_mpmf: pandas.DataFrame
        """
        super().__init__(optim, results, df_mpmf)
        self.gene2name = {row['label']: row.get('name')
                          for gpid, row in self.optim.m_dict['fbcGeneProducts'].iterrows()}

    def get_fluxes(self, solution):
        """Collect metabolic isoreaction fluxes for a single optimization solution.

        Excluding flux information for variables starting with 'V_'

        Fluxes are rescaled to mmol/gDWh
        Note: some synthesis and degradation fluxes are in µmol/gDWh

        :param solution: optimization solution
        :type solution: :class:`Solution`
        :return: reaction fluxes with addition reaction data added
        :rtype: pandas.DataFrame
        """
        # remove 'R_' prefix to align with CobraPy reaction ids
        pf_prod = re.sub(f'^{pf.R_}', '', pf.R_PROD_)
        pf_degr = re.sub(f'^{pf.R_}', '', pf.R_DEGR_)

        fluxes = {}
        for rid, flux in solution.fluxes.items():
            if re.match(pf.V_, rid) is None:
                # check for macromolecule synthesis/degradation fluxes that might be scaled
                if re.match(pf_prod, rid) or re.match(pf_degr, rid):
                    if re.match(pf_prod, rid):
                        mm_id = re.sub(pf_prod, '', rid)
                    else:
                        mm_id = re.sub(pf_degr, '', rid)
                    scale = self.optim.df_mm_data.at[mm_id, 'scale']
                    mmol_per_gdwh = flux / scale
                else:
                    # metabolic reaction fluxes are considered to be in units of mmol/gDWh
                    mmol_per_gdwh = flux
                rdata = self.optim.rdata[rid]
                fluxes[rid] = [rdata['reaction_str'], rdata['net_rid'], rdata['gpr'],
                               mmol_per_gdwh, abs(mmol_per_gdwh)]
        cols = ['reaction_str', 'net_rid', 'gpr', 'mmol_per_gDWh', 'abs mmol_per_gDWh']
        df_fluxes = pd.DataFrame(fluxes.values(), index=list(fluxes), columns=cols)
        df_fluxes.index.name = 'reaction'
        return df_fluxes

    def get_net_fluxes(self, solution):
        """Extract net metabolic reaction fluxes for a given optimization solution.

        Combinding fluxes of isoreactions on reaction level.
        Including forward/reverse fluxes
        rescale macromolecule synthesis and reaction fluxes to mmol/gDWh

        Excluding flux information for variables starting with 'V_'

        :param solution: optimization solution
        :type solution: :class:`Solution`
        :return: table with predicted net reaction fluxes
        :rtype: pandas.DataFrame
        """
        # remove 'R_' prefix to align with CobraPy reaction ids
        pf_prod = re.sub(f'^{pf.R_}', '', pf.R_PROD_)
        pf_degr = re.sub(f'^{pf.R_}', '', pf.R_DEGR_)

        net_fluxes = defaultdict(float)
        for rid, val in solution.fluxes.items():
            if re.match(pf.V_, rid) is None:
                # check for macromolecule synthesis/degradation fluxes that might be scaled
                if re.match(pf_prod, rid) or re.match(pf_degr, rid):
                    if re.match(pf_prod, rid):
                        mm_id = re.sub(pf_prod, '', rid)
                    else:
                        mm_id = re.sub(pf_degr, '', rid)
                    scale = self.optim.df_mm_data.at[mm_id, 'scale']
                    net_fluxes[rid] = val / scale
                else:
                    # collaps any iso reactions, and combine forward/reverse reactions
                    net_rid = re.sub(r'_iso\d*', '', rid)
                    net_rid = re.sub('_REV$', '', net_rid)
                    if re.search('_REV', rid):
                        net_fluxes[net_rid] -= val
                    else:
                        net_fluxes[net_rid] += val

        # add reaction string and gene product relations
        net_flux_data = {}
        for rid, mmol_per_gdwh in net_fluxes.items():
            rdata = self.optim.net_rdata.get(rid)
            if rdata:
                net_flux_data[rid] = [rdata['reaction_str'], rdata['gpr'], mmol_per_gdwh, abs(mmol_per_gdwh)]
            else:
                net_flux_data[rid] = [None, None, mmol_per_gdwh, abs(mmol_per_gdwh)]

        cols = ['reaction_str', 'gpr', 'mmol_per_gDWh', 'abs mmol_per_gDWh']
        df_net_fluxes = pd.DataFrame(net_flux_data.values(), index=list(net_flux_data), columns=cols)
        df_net_fluxes.index.name = 'rid'
        return df_net_fluxes

    def get_predicted_protein_data(self, solution):
        """Determine protein levels for a given optimization solution.

        Enzyme, process machinery and dummy protein concentrations
        in µmol/gDW, mg/gDW and pmf (protein mass fraction)

        pmf is calculated by dividing individual protein mass by total protein
        mass for that condition.

        :param solution: optimization solution
        :type solution: :class:`Solution`
        :return: table with predicted protein concentrations in µmol/gDW, mg/gDW and mg/gP
        :rtype: pandas.DataFrame
        """
        protein_mmol_per_gdw = defaultdict(float)
        for var_id, conc in solution.fluxes.items():
            # iterate through all enzyme and process machine concentrations
            if re.match(pf.V_EC_, var_id) or re.match(pf.V_PMC_, var_id):
                scale = self.optim.df_enz_data.at[var_id, 'scale']
                for gp_id, stoic in self.optim.enz_mm_composition[var_id].items():
                    # exclude any RNAs (e.g. in ribosome)
                    if self.optim.df_mm_data.at[gp_id, 'uniprot']:
                        protein_mmol_per_gdw[gp_id] += stoic * conc / scale
            # include protein concentrations from concentration targets (proteins not included in enzymes)
            elif re.match(pf.V_TMMC_, var_id):
                gp_id = re.sub(pf.V_TMMC_, '', var_id)
                if self.optim.df_mm_data.at[gp_id, 'type'] == 'proteins':
                    scale = self.optim.df_mm_data.at[gp_id, 'scale']
                    protein_mmol_per_gdw[gp_id] += conc / scale

        prot_data = []
        for gene, mmol_per_gdw in protein_mmol_per_gdw.items():
            uniprot = self.optim.df_mm_data.at[gene, 'uniprot']
            mw_kda = self.optim.df_mm_data.at[gene, 'mw_kDa']
            mg_per_gdw = mmol_per_gdw * mw_kda * 1000.0
            gene_name = self.gene2name.get(gene)
            prot_data.append([gene, uniprot, gene_name, mw_kda, mmol_per_gdw * 1000.0, mg_per_gdw])
        cols = ['gene', 'uniprot', 'gene_name', 'mw_kDa', 'µmol_per_gDW', 'mg_per_gDW']
        df_prot_data = pd.DataFrame(prot_data, columns=cols).set_index('gene')
        total_gp_per_gdw = df_prot_data['mg_per_gDW'].sum() / 1000.0
        df_prot_data['mg_per_gP'] = df_prot_data['mg_per_gDW'] / total_gp_per_gdw
        return df_prot_data

    def collect_protein_results(self, units='mg_per_gP'):
        """Extract predicted protein concentrations from optimization results.

        Information columns are incorporated to provide context and facilitate data filtration.

        :param str units: concentration units: 'mg_per_gP', 'mg_per_gDW' or 'µmol_per_gDW' (default: 'mg_per_gP')
        :return: table with predicted enzyme concentrations
        :rtype: pandas.DataFrame
        """
        supported_units = {'mg_per_gP', 'mg_per_gDW', 'µmol_per_gDW'}
        if units not in supported_units:
            print(f'{units} not support, select any of {supported_units}')
            return pd.DataFrame()
        else:
            return self._collect_protein_results(units)

    def get_predicted_enzyme_usage(self, solution):
        enz_usage = {}
        for vid, flux in solution.fluxes.items():
            if re.match(pf.V_EC_, vid) or re.match(pf.V_PMC_, vid):
                mw_kda = self.optim.df_enz_data.at[vid, 'mw_kDa']
                scale = self.optim.df_enz_data.at[vid, 'scale']
                mmf = flux * mw_kda / scale * 1000.0
                conc_umol_gdw = flux / scale * 1000.0
                rid = re.sub(f'^({pf.V_EC_}|{pf.V_PMC_})', '', vid)

                # collect reaction related information
                if f'R_{rid}' in self.optim.rdata:
                    rdata = self.optim.rdata[f'R_{rid}']
                    gpr = rdata['gpr']
                    reaction_str = rdata['reaction_str']
                elif rid in self.optim.rdata:
                    rdata = self.optim.rdata[rid]
                    gpr = rdata['gpr']
                    reaction_str = rdata['reaction_str']
                else:
                    gpr = None
                    reaction_str = None
                enz_usage[rid] = [mw_kda, reaction_str, gpr, conc_umol_gdw, mmf]

        cols = ['mw_kDa', 'reaction_str', 'gpr', 'µmol_per_gDW', 'mg_per_gDW']
        df_enz_usage = pd.DataFrame(enz_usage.values(), index=list(enz_usage), columns=cols)
        return df_enz_usage

    def collect_enzyme_results(self, units='mg_per_gDW'):
        """Extract predicted enzyme concentrations from optimization results.

        Information columns are incorporated to provide context and facilitate data filtration.

        :param str units: concentration units: 'mg_per_gDW' or 'µmol_per_gDW' (default: 'mg_per_gDW')
        :return: table with predicted enzyme concentrations
        :rtype: pandas.DataFrame
        """
        supported_units = {'mg_per_gDW', 'µmol_per_gDW'}
        if units not in supported_units:
            print(f'{units} not support, select any of {supported_units}')
            return None

        df_enzymes = None
        info_cols = ['mw_kDa', 'reaction_str', 'gpr']
        n_info_cols = len(info_cols)
        for condition, solution in self.results.items():
            df = self.get_predicted_enzyme_usage(solution)
            if df_enzymes is None:
                df_enzymes = df[info_cols + [units]].copy()
            else:
                df_enzymes = pd.concat([df_enzymes, df[units]], axis=1)
            df_enzymes.rename(columns={units: f'{condition}'}, inplace=True)
        mean = df_enzymes.iloc[:, n_info_cols:].mean(axis=1).values
        stdev = df_enzymes.iloc[:, n_info_cols:].std(axis=1).values
        df_enzymes.insert(n_info_cols, f'mean {units}', mean)
        df_enzymes.insert(n_info_cols + 1, 'stdev', stdev)
        df_enzymes.index.name = 'enzyme'
        df_enzymes.sort_values(by=f'mean {units}', ascending=False, inplace=True)
        return df_enzymes

    def get_predicted_rna_usage(self, solution):
        rna_mmol_per_gdw = defaultdict(float)
        for var_id, conc in solution.fluxes.items():
            # get rRNAs from process machines concentrations
            if re.match(pf.V_PMC_, var_id):
                scale = self.optim.df_enz_data.at[var_id, 'scale']
                for rna_id, stoic in self.optim.enz_mm_composition[var_id].items():
                    # exclude any proteins (which have uniprot)
                    if self.optim.df_mm_data.at[rna_id, 'uniprot'] is None:
                        rna_mmol_per_gdw[rna_id] += stoic * conc / scale
            # get tRNAs, mRNA from target macromolecule concentrations
            elif re.match(pf.V_TMMC_, var_id) and 'rna' in var_id.lower():
                rna_id = re.sub(f'^{pf.V_TMMC_}', '', var_id)
                scale = self.optim.df_mm_data.at[rna_id, 'scale']
                rna_mmol_per_gdw[rna_id] += conc / scale

        rna_data = {}
        for rna_id, mmol_per_gdw in rna_mmol_per_gdw.items():
            mw_kda = self.optim.df_mm_data.at[rna_id, 'mw_kDa']
            mg_per_gdw = mmol_per_gdw * mw_kda * 1000.0
            rna_data[rna_id] = [mw_kda, mmol_per_gdw * 1000.0, mg_per_gdw]

        cols = ['mw_kDa', 'µmol_per_gDW', 'mg_per_gDW']
        df_rna_data = pd.DataFrame(rna_data.values(), index=list(rna_data), columns=cols)
        total_grna_per_gdw = df_rna_data['mg_per_gDW'].sum() / 1000.0
        df_rna_data['mg_per_gRNA'] = df_rna_data['mg_per_gDW'] / total_grna_per_gdw
        df_rna_data.index.name = 'gene'
        return df_rna_data

    def collect_rna_results(self, units='mg_per_gDW'):
        """Extract predicted RNA concentrations from optimization results.

        Information columns are incorporated to provide context and facilitate data filtration.

        :param str units: concentration units: 'mg_per_gDW' or 'µmol_per_gDW' (default: 'mg_per_gDW')
        :return: table with predicted RNA concentrations
        :rtype: pandas.DataFrame
        """
        supported_units = {'mg_per_gDW', 'µmol_per_gDW'}
        if units not in supported_units:
            print(f'{units} not support, select any of {supported_units}')
            return None

        df_rnas = None
        info_cols = ['mw_kDa']
        n_info_cols = 1
        for condition, solution in self.results.items():
            df = self.get_predicted_rna_usage(solution)
            if df_rnas is None:
                df_rnas = df[info_cols + [units]].copy()
            else:
                df_rnas = pd.concat([df_rnas, df[units]], axis=1)
            df_rnas.rename(columns={units: f'{condition}'}, inplace=True)
        mean = df_rnas.iloc[:, n_info_cols:].mean(axis=1).values
        stdev = df_rnas.iloc[:, n_info_cols:].std(axis=1).values
        df_rnas.insert(n_info_cols, f'mean {units}', mean)
        df_rnas.insert(n_info_cols + 1, 'stdev', stdev)
        df_rnas.index.name = 'rna'
        df_rnas.sort_values(by=f'mean {units}', ascending=False, inplace=True)
        return df_rnas

    @staticmethod
    def get_occupancy(solution):
        """Get compartment occupancy for single optimization solution.

        Calculated based on slack variables for density constraints
        Based on compartment capacity (added to solution data) the
        % protein occupancy is calculated

        :param solution: optimization solution
        :type solution: :class:`Solution`
        :return: compartment occupancy
        :rtype: pandas.DataFrame
        """
        occupancy = {}
        for var_id, val in solution.fluxes.items():
            if re.match(pf.V_SLACK_, var_id):
                cid = re.sub(f'^{pf.V_SLACK_}', '', var_id)
                capacity = solution.density_constraints[f'{pf.C_D_}{cid}']
                occ = capacity - val
                occupancy[cid] = [occ / capacity, occ, capacity]
        cols = ['utilization', 'used mmolAA_per_gDW', 'capacity mmolAA_per_gDW']
        df_occ = pd.DataFrame(occupancy.values(), index=list(occupancy), columns=cols)
        df_occ.index.name = 'cid'
        return df_occ

    def collect_density_results(self, capacity=False):
        """Extract predicted density levels from optimization results.

        With `capacity` set to true the density levels are reported in percentages.
        Information columns are incorporated to provide context and facilitate data filtration.

        :param bool capacity: report capacity levels (default: False)
        :return: table with predicted density levels
        :rtype: pandas.DataFrame
        """
        col_name = 'capacity mmolAA_per_gDW' if capacity else 'utilization'
        df_occupancy = None
        info_cols = []
        n_info_cols = len(info_cols)
        for condition, solution in self.results.items():
            df = self.get_occupancy(solution)
            if df_occupancy is None:
                df_occupancy = df[info_cols + [col_name]].copy()
            else:
                df_occupancy = pd.concat([df_occupancy, df[col_name]], axis=1)
            df_occupancy.rename(columns={col_name: f'{condition}'}, inplace=True)
        mean = df_occupancy.iloc[:, n_info_cols:].mean(axis=1).values
        stdev = df_occupancy.iloc[:, n_info_cols:].std(axis=1).values
        df_occupancy.insert(0, f'mean {col_name}', mean)
        df_occupancy.insert(1, 'stdev', stdev)
        df_occupancy.sort_values(by=f'mean {col_name}', ascending=False, inplace=True)
        df_occupancy.index.name = 'cid'
        return df_occupancy
